/* sane - Scanner Access Now Easy.

   Copyright (C) 2000 Mustek.
   Maintained by Tom Wang <tom.wang@mustek.com.tw>

   Updates (C) 2001 by Henning Meier-Geinitz.

   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file implements a SANE backend for Mustek 1200UB flatbed scanners.  */

#include "mustek_usb_high.h"
#include "mustek_usb_mid.c"

/* ------------------------ calibration functions ------------------------- */

static SANE_Byte GrayMap[8] = 
{
  0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
};

static inline double
FiltLower (int *pSortBuffer, SANE_Word TotalCount, SANE_Word FiltCount)
{
  SANE_Word Bound = TotalCount - 1;
  SANE_Word LeftCount = TotalCount - FiltCount;
  int Temp = 0;
  SANE_Word i, j;
  int Sum = 0;

  for (i = 0; i < Bound; i++)
    {
      for (j = 0; j < Bound - i; j++)
	{
	  if (pSortBuffer[j + 1] > pSortBuffer[j])
	    {
	      Temp = pSortBuffer[j];
	      pSortBuffer[j] = pSortBuffer[j + 1];
	      pSortBuffer[j + 1] = Temp;
	    }
	}
    }
  for (i = 0; i < LeftCount; i++)
    Sum += pSortBuffer[i];
  return (double) Sum;
}

SANE_Status
usb_high_cal_MD12_init (CalMD12 * cal, SANE_Byte type, SANE_Word wTargetWhite,
			SANE_Word wTargetDark)
{
  DBG (5, "usb_high_cal_MD12_init: start, cal=%p, type=%d, wTargetWhite=%d "
       "wTargetDark=%d\n", cal, type, wTargetWhite, wTargetDark);
  cal->isPrepared = SANE_FALSE;
  cal->pKWhite = NULL;
  cal->pKDark = NULL;
  /* Working Buffer */
  cal->pWhiteLine = NULL;
  cal->pDarkLine = NULL;
  cal->pWhiteBuffer = NULL;
  /* Necessary Parameters */
  cal->wKWhiteLevel = 240 << 8;
  cal->wKDarkLevel = 0;
  cal->wThreshold = 2048;
  cal->wMajorAver = 0;
  cal->wMinorAver = 0;
  cal->wFilter = 0;
  cal->wWhiteNeeded = 0;
  cal->wDarkNeeded = 0;
  cal->wMaxWidth = 0;
  cal->wWidth = 100;
  cal->pGammaTable = NULL;
  cal->nCalibratorType = type;
  cal->wKWhiteLevel = wTargetWhite / 16;
  cal->wKDarkLevel = 0;
  DBG (5, "usb_high_cal_MD12_init: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12_exit (CalMD12 * cal)
{
  DBG (5, "usb_high_cal_MD12_exit: start\n");

  if (!cal)
    {
      DBG (3, "usb_high_cal_MD12_exit: cal == NULL\n");
      return SANE_STATUS_INVAL;
    }

  if (!cal->isPrepared)
    {
      DBG (3, "usb_high_cal_MD12_exit: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }
  DBG (5, "usb_high_cal_MD12_exit: 1\n");

  if (cal->pKDark)
    {
      free (cal->pKDark);
    }
  cal->pKDark = NULL;
  DBG (5, "usb_high_cal_MD12_exit: 2\n");
  if (cal->pKWhite)
    {
      free (cal->pKWhite);
    }
  cal->pKWhite = NULL;
  DBG (5, "usb_high_cal_MD12_exit: 3\n");

  cal->isPrepared = SANE_FALSE;
  DBG (5, "usb_high_cal_MD12_exit: 4\n");
  DBG (5, "usb_high_cal_MD12_exit: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12_EmbedGamma (CalMD12 * cal, SANE_Word * pGammaTable)
{
  DBG (5, "usb_high_cal_MD12_EmbedGamma: start\n");
  cal->pGammaTable = pGammaTable;
  DBG (5, "usb_high_cal_MD12_EmbedGamma: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12_Prepare (CalMD12 * cal, SANE_Word wMaxWidth)
{
  DBG (5, "usb_high_cal_MD12_Parepare: start\n");

  if (cal->isPrepared)
    {
      DBG (3, "usb_high_cal_MD12_Parepare: isPrepared\n");
      return SANE_STATUS_INVAL;
    }

  if (cal->pKWhite)
    {
      free (cal->pKWhite);
    }
  cal->pKWhite = (SANE_Word *) malloc (wMaxWidth * sizeof (SANE_Word));
  if (!cal->pKWhite)
    return SANE_STATUS_NO_MEM;

  if (cal->pKDark)
    {
      free (cal->pKDark);
    }
  cal->pKDark = (SANE_Word *) malloc (wMaxWidth * sizeof (SANE_Word));
  if (!cal->pKDark)
    return SANE_STATUS_NO_MEM;

  cal->wMaxWidth = wMaxWidth;

  cal->isPrepared = SANE_TRUE;

  DBG (5, "usb_high_cal_MD12_Parepare: exit\n");
  return SANE_STATUS_GOOD;
}


SANE_Status
usb_high_cal_MD12_Setup (CalMD12 * cal, SANE_Word wMojorAver,
			 SANE_Word wMinorAver, SANE_Word wFilter, 
			 SANE_Word wWidth, SANE_Word * wWhiteNeeded,
			 SANE_Word * wDarkNeeded)
{
  int i;

  DBG (5, "usb_high_cal_MD12_Setup: start\n");

  if (!cal->isPrepared)
    {
      DBG (3, "usb_high_cal_MD12_Setup: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }
  if (wMojorAver == 0)
    {
      DBG (3, "usb_high_cal_MD12_Setup: wMojorAver==0\n");
      return SANE_STATUS_INVAL;
    }
  if (wMinorAver == 0)
    {
      DBG (3, "usb_high_cal_MD12_Setup: wMinorAver==0\n");
      return SANE_STATUS_INVAL;
    }
  if (wWidth > cal->wMaxWidth)
    {
      DBG (3, "usb_high_cal_MD12_Setup: wWidth>wMaxWidth\n");
      return SANE_STATUS_INVAL;
    }

  cal->wMajorAver = wMojorAver;
  cal->wMinorAver = wMinorAver;
  cal->wFilter = wFilter;
  cal->wWidth = wWidth;
  cal->wWhiteNeeded = wMojorAver * 16 + wFilter;
  cal->wDarkNeeded = wMojorAver * 16;
  *wWhiteNeeded = cal->wWhiteNeeded;
  *wDarkNeeded = cal->wDarkNeeded;

  if (cal->pWhiteLine)
    {
      free (cal->pWhiteLine);
    }
  cal->pWhiteLine =
    (double *) malloc (cal->wWidth * sizeof (double));
  if (!cal->pWhiteLine)
    return SANE_STATUS_NO_MEM;

  if (cal->pDarkLine)
    {
      free (cal->pDarkLine);
    }
  cal->pDarkLine =
    (double *) malloc (cal->wWidth * sizeof (double));
  if (!cal->pDarkLine)
    return SANE_STATUS_NO_MEM;

  for (i = 0; i < cal->wWidth; i++)
    {
      cal->pWhiteLine[i] = 0.0;
      cal->pDarkLine[i] = 0.0;
    }

  if (cal->pWhiteBuffer)
    {
      free (cal->pWhiteBuffer);
    }
  cal->pWhiteBuffer =
    (int *) malloc (cal->wWhiteNeeded * cal->wWidth * sizeof (int));
  if (!cal->pWhiteBuffer)
    return SANE_STATUS_NO_MEM;

  for (i = 0; i < cal->wWhiteNeeded * cal->wWidth; i++)
    {
      *(cal->pWhiteBuffer + i) = 0;
    }

  return SANE_STATUS_GOOD;
  DBG (5, "usb_high_cal_MD12_Setup: start\n");
}

SANE_Status
usb_high_cal_MD12_EvalueWhite (CalMD12 * cal, double dFactor)
{
  /* Caculate pWhiteLine */
  double LoopDivision;
  double Aver;
  int *pSortBuffer;
  SANE_Word i, j;

  DBG (5, "usb_high_cal_MD12_EvalueWhite: start\n");
  LoopDivision = (double) (cal->wMajorAver * cal->wMinorAver);
  pSortBuffer = (int *) malloc (cal->wWhiteNeeded * sizeof (int));
  if (!pSortBuffer)
    return SANE_STATUS_NO_MEM;
  
  if (cal->pWhiteBuffer == NULL)
    {
      DBG (3, "usb_high_cal_MD12_EvalueWhite: pWhiteBuffer==NULL\n");
      return SANE_STATUS_NO_MEM;
    }

  for (i = 0; i < cal->wWidth; i++)
    {
      for (j = 0; j < cal->wWhiteNeeded; j++)
	{
	  *(pSortBuffer + j) =
	    *(cal->pWhiteBuffer + j * cal->wWidth + i);
	}
      Aver =
	FiltLower (pSortBuffer, cal->wWhiteNeeded,
		   cal->wFilter) * dFactor / LoopDivision;
      if (Aver >= 4096.0)
	cal->pWhiteLine[i] = 4095.9999;
      else if (Aver < 0.0)
	cal->pWhiteLine[i] = 0.0;
      else
	cal->pWhiteLine[i] = Aver;
    }
  free (pSortBuffer);
  pSortBuffer = NULL;
  free (cal->pWhiteBuffer);
  cal->pWhiteBuffer = NULL;
  DBG (5, "usb_high_cal_MD12_EvalueWhite: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12_EvalueDark (CalMD12 * cal, double dFactor)
{
  SANE_Word i;
  double LoopDivision;

  DBG (5, "usb_high_cal_MD12_EvalueWhite: exit\n");
  /* Caculate pDarkLine */
  dFactor *= 16.0;
  LoopDivision = (double) (cal->wMajorAver * cal->wMinorAver);
  for (i = 0; i < cal->wWidth; i++)
    {
      cal->pDarkLine[i] /= LoopDivision;
      cal->pDarkLine[i] -= dFactor;
      if (cal->pDarkLine[i] < 0.0)
	cal->pDarkLine[i] = 0.0;
    }
  DBG (5, "usb_high_cal_MD12_EvalueDark: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12_EvalueCalibrator (CalMD12 * cal)
{
  int Aver = 0;
  SANE_Word i;

  DBG (5, "usb_high_cal_MD12_EvalueCalibrator: start\n");
  if (cal->pWhiteLine == NULL)
    {
      DBG (3, "usb_high_cal_MD12_EvalueCalibrator: pWhiteLine==NULL\n");
      return SANE_FALSE;
    }
  if (cal->pDarkLine == NULL)
    {
      DBG (3, "usb_high_cal_MD12_EvalueCalibrator: pDarkLine==NULL\n");
      return SANE_FALSE;
    }

  for (i = 0; i < cal->wWidth; i++)
    {
      Aver = (int) (cal->pWhiteLine[i]) - (int) (cal->pDarkLine[i]);
      if (Aver <= 0)
	Aver = 1;
      else if (Aver >= 4096)
	Aver = 4095;
      cal->pKWhite[i] = (SANE_Word) (Aver);
      cal->pKDark[i] = (SANE_Word) (cal->pDarkLine[i]);
    }
  free (cal->pDarkLine);
  cal->pDarkLine = NULL;
  free (cal->pWhiteLine);
  cal->pWhiteLine = NULL;

  DBG (5, "usb_high_cal_MD12_EvalueCalibrator: start\n");
  return SANE_STATUS_GOOD;
}

/* virtual function switcher */
SANE_Status
usb_high_cal_MD12_FillInWhite (CalMD12 * cal, SANE_Word wMajor,
			       SANE_Word wMinor, void * pWhitePattern)
{
  DBG (5, "usb_high_cal_MD12_FillInWhite: start\n");
  switch (cal->nCalibratorType)
    {
    case I8O8RGB:
    case I8O8MONO:
      return usb_high_cal_MD12i8o8_FillInWhite (cal, wMajor, wMinor,
						pWhitePattern);
      break;
    case I4O1MONO:
      return usb_high_cal_MD12i4o1_FillInWhite (cal, wMajor, wMinor,
						pWhitePattern);
      break;
    }
  DBG (5, "usb_high_cal_MD12_FillInWhite: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12_FillInDark (CalMD12 * cal, SANE_Word wMajor,
			     SANE_Word wMinor, void * pDarkPattern)
{
  DBG (5, "usb_high_cal_MD12_FillInDark: start\n");
  switch (cal->nCalibratorType)
    {
    case I8O8RGB:
    case I8O8MONO:
      return usb_high_cal_MD12i8o8_FillInDark (cal, wMajor, wMinor,
					      pDarkPattern);
      break;
    case I4O1MONO:
      return usb_high_cal_MD12i4o1_FillInDark (cal, wMajor, wMinor,
					      pDarkPattern);
      break;
    }
  DBG (5, "usb_high_cal_MD12_FillInDark: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12_Calibrate (CalMD12 * cal, void * pSrc,
			    void * pTarget)
{
  DBG (5, "usb_high_cal_MD12_Calibrate: start\n");
  switch (cal->nCalibratorType)
    {
    case I8O8RGB:
      return usb_high_cal_MD12i8o8Rgb_Calibrate (cal, pSrc, pTarget);
      break;
    case I8O8MONO:
      return usb_high_cal_MD12i8o8Mono_Calibrate (cal, pSrc, pTarget);
      break;
    case I4O1MONO:
      return usb_high_cal_MD12i4o1_Calibrate (cal, pSrc, pTarget);
      break;
    }
  DBG (5, "usb_high_cal_MD12_Calibrate: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12i8o8_FillInWhite (CalMD12 * cal, SANE_Word wMajor,
				   SANE_Word Minor, void * pWhitePattern)
{
  SANE_Byte * pPattern;
  SANE_Word j;

  pPattern = (SANE_Byte *) pWhitePattern;

  DBG (5, "usb_high_cal_MD12i8o8_FillInWhite: start, Minor=%d\n", Minor);
  if (!cal->isPrepared)
    {
      DBG (3, "usb_high_cal_MD12i8o8_FillInWhite: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }
  if (cal->wWhiteNeeded == 0)
    {
      DBG (3, "usb_high_cal_MD12i8o8_FillInWhite: wWhiteNeeded==0\n");
       return SANE_STATUS_INVAL;
    }

  for (j = 0; j < cal->wWidth; j++)
    {
      *(cal->pWhiteBuffer + wMajor * cal->wWidth + j) +=
	(int) (pPattern[j]);
    }
  DBG (5, "usb_high_cal_MD12i8o8_FillInWhite: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12i8o8_FillInDark (CalMD12 * cal, SANE_Word wMajor,
				  SANE_Word Minor, void * pDarkPattern)
{
  SANE_Byte * pPattern = (SANE_Byte *) pDarkPattern;
  SANE_Word j;

  DBG (5, "usb_high_cal_MD12i8o8_FillInDark: start, wMajor=%d, Minor=%d\n",
       wMajor, Minor);
  if (!cal->isPrepared)
    {
      DBG (3, "usb_high_cal_MD12i8o8_FillInDark: !isPrepared\n");
      return SANE_FALSE;
    }
  if (cal->wDarkNeeded == 0)
    {
      DBG (3, "usb_high_cal_MD12i8o8_FillInDark: wDarkNeeded==0\n");
      return SANE_FALSE;
    }

  for (j = 0; j < cal->wWidth; j++)
    {
      cal->pDarkLine[j] += (double) (pPattern[j]);
    }
  DBG (5, "usb_high_cal_MD12i8o8_FillInDark: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12i4o1_FillInWhite (CalMD12 * cal, SANE_Word wMajor,
				   SANE_Word Minor, void * pWhitePattern)
{
  SANE_Byte * pPattern;
  SANE_Word j = 0;

  pPattern = (SANE_Byte *) pWhitePattern;

  DBG(5, "usb_high_cal_MD12i4o1_FillInWhite: Minor=%d\n", Minor);
  if (!cal->isPrepared)
    {
      DBG (3, "usb_high_cal_MD12i4o1_FillInWhite: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }
  if (cal->wWhiteNeeded == 0)
    {
      DBG (3, "usb_high_cal_MD12i4o1_FillInWhite: wWhiteNeeded==0\n");
      return SANE_STATUS_INVAL;
    }

  while (j < cal->wWidth)
    {
      *(cal->pWhiteBuffer + wMajor * cal->wWidth + j) +=
	(int) (*(pPattern) & 0xf0);
      j++;
      if (j >= cal->wWidth)
	break;
      *(cal->pWhiteBuffer + wMajor * cal->wWidth + j) +=
	(int) ((SANE_Byte) (*(pPattern++) << 4));
      j++;
    }
  DBG (5, "usb_high_cal_MD12i8o8_FillInWhite: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12i4o1_FillInDark (CalMD12 * cal, SANE_Word wMajor,
				  SANE_Word Minor, void * pDarkPattern)
{
  SANE_Byte * pPattern;
  SANE_Word j = 0;

  pPattern = (SANE_Byte *) pDarkPattern;

  DBG (5,"usb_high_cal_MD12i4o1_FillInDark: start, wMajor=%d, Minor=%d\n",
       wMajor, Minor); 
  if (!cal->isPrepared)
    {
      DBG (3, "usb_high_cal_MD12i4o1_FillInDark: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }
  if (cal->wDarkNeeded == 0)
    {
      DBG (5,"usb_high_cal_MD12i4o1_FillInDark: wDarkNeeded==0\n");
      return SANE_STATUS_INVAL;
    }

  while (j < cal->wWidth)
    {
      cal->pDarkLine[j++] += (double) (*(pPattern) & 0xf0);
      if (j >= cal->wWidth)
	break;
      cal->pDarkLine[j++] += (double) ((SANE_Byte) (*(pPattern++) << 4));
    }
  DBG (5,"usb_high_cal_MD12i4o1_FillInDark: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12i8o8Mono_Calibrate (CalMD12 * cal, void * pSrc,
				     void * pTarget)
{
  SANE_Byte *pGraySrc;
  SANE_Byte *pGrayTarget;
  int Base = 0;
  SANE_Word Value = 0;
  SANE_Word i;

  DBG(5, "usb_high_cal_MDi8o8Mono_Calibrate: start\n");

  pGraySrc = (SANE_Byte *) pSrc;
  pGrayTarget = (SANE_Byte *) pTarget;

  if (cal->pGammaTable == NULL)
    {
      SANE_Word wKWhiteLevel = cal->wKWhiteLevel >> 4;
      for (i = 0; i < cal->wWidth; i++)
	{
	  Base =
	    (int) ((SANE_Word) (pGraySrc[i]) << 4) - (int) (cal->pKDark[i]);
	  if (Base < 0)
	    Base = 0;
	  Value = ((SANE_Word) (Base) * wKWhiteLevel) / cal->pKWhite[i];
	  if (Value > 0x00ff)
	    Value = 0x00ff;
	  pGrayTarget[i] = (SANE_Byte) (Value);
	}
    }
  else
    {
      for (i = 0; i < cal->wWidth; i++)
	{
	  Base =
	    (int) ((SANE_Word) (pGraySrc[i]) << 4) - (int) (cal->pKDark[i]);
	  if (Base < 0)
	    Base = 0;
	  Value =
	    ((SANE_Word) (Base) * cal->wKWhiteLevel) / cal->pKWhite[i];
	  if (Value > 0x0fff)
	    Value = 0x0fff;
	  pGrayTarget[i] = (SANE_Byte) (cal->pGammaTable[Value]);
	}
    }
  DBG(5, "usb_high_cal_MDi8o8Mono_Calibrate: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12i8o8Rgb_Calibrate (CalMD12 * cal, void * pSrc,
				   void * pTarget)
{
  SANE_Byte *pGraySrc;
  SANE_Byte *pRgbTarget;
  int Base = 0;
  SANE_Word Value = 0;
  SANE_Word i;

  DBG (5, "usb_high_cal_MDi8o8Rgb_Calibrate: start\n");
  pGraySrc = (SANE_Byte *) pSrc;
  pRgbTarget = (SANE_Byte *) pTarget;

  if (cal->pGammaTable == NULL)
    {
      SANE_Word wKWhiteLevel = cal->wKWhiteLevel >> 4;
      for (i = 0; i < cal->wWidth; i++)
	{
	  Base =
	    (int) ((SANE_Word) (pGraySrc[i]) << 4) - (int) (cal->pKDark[i]);
	  if (Base < 0)
	    Base = 0;
	  Value = ((SANE_Word) (Base) * wKWhiteLevel) / cal->pKWhite[i];
	  if (Value > 0x00ff)
	    Value = 0x00ff;
	  *pRgbTarget = (SANE_Byte) (Value);
	  pRgbTarget += 3;
	}
    }
  else
    {
      for (i = 0; i < cal->wWidth; i++)
	{
	  Base =
	    (int) ((SANE_Word) (pGraySrc[i]) << 4) - (int) (cal->pKDark[i]);
	  if (Base < 0)
	    Base = 0;
	  Value =
	    ((SANE_Word) (Base) * cal->wKWhiteLevel) / cal->pKWhite[i];
	  if (Value > 0x0fff)
	    Value = 0x0fff;
	  *(pRgbTarget) = (SANE_Byte) (cal->pGammaTable[Value]);
	  pRgbTarget += 3;
	}
    }
  DBG (5, "usb_high_cal_MDi8o8Rgb_Calibrate: start\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_cal_MD12i4o1_Calibrate (CalMD12 * cal, void * pSrc, void * pTarget)
{
  SANE_Byte *p4BitSrc;
  SANE_Byte *p1BitTarget;
  int Base = 0;
  SANE_Word Value = 0;
  SANE_Word j = 0;
  int Count = 0;

  DBG (5,"usb_high_cal_MD12i4o1_Calibrate: start\n");
  p4BitSrc = (SANE_Byte *) pSrc;
  p1BitTarget = (SANE_Byte *) pTarget;

  *p1BitTarget = 0;
  while (j < cal->wWidth)
    {
      Base =
	(int) ((SANE_Word) (*p4BitSrc & 0xf0) << 4) - (int) (cal->pKDark[j]);
      if (Base < 0)
	Base = 0;
      Value = ((SANE_Word) (Base) * cal->wKWhiteLevel) / cal->pKWhite[j];
      if (Value > 0x0fff)
	Value = 0x0fff;
      if (Value >= cal->wThreshold)
	*(p1BitTarget) |= GrayMap[Count];
      Count++;
      j++;
      if (j >= cal->wWidth)
	break;
      Base =
	(int) ((SANE_Word) (*(p4BitSrc++) & 0x0f) << 8) -
	(int) (cal->pKDark[j]);
      if (Base < 0)
	Base = 0;
      Value = ((SANE_Word) (Base) * cal->wKWhiteLevel) / cal->pKWhite[j];
      if (Value > 0x0fff)
	Value = 0x0fff;
      if (Value >= cal->wThreshold)
	*(p1BitTarget) |= GrayMap[Count];
      Count++;
      if (Count >= 8)
	{
	  p1BitTarget++;
	  *p1BitTarget = 0;
	  Count = 0;
	}
      j++;
    }
  DBG (5,"usb_high_cal_MD12i4o1_Calibrate: exit\n");
  return SANE_STATUS_GOOD;
}


/* --------------------------- scan functions ----------------------------- */


SANE_Status
usb_high_scan_init (Mustek_Usb_Device * dev)
{
  SANE_Status status;

  DBG (5, "usb_high_scan_init: start\n");

  dev->INI_MinExploseTime = A4CIS_MinExploseTime;
  dev->INI_300SkipsPerRow = A4CIS_SkipsPerRow;
  dev->INI_600SkipsPerRow = A4CIS_SkipsPerRow;
  dev->INI_JLines = A4CIS_JLines;
  dev->INI_KLines = A4CIS_KLines;
  dev->INI_KFilter = A4CIS_KFilter;
  dev->INI_KLoops = A4CIS_KLoops;
  dev->INI_PixelRateLines = 50;
  dev->INI_PixelRateFilts = 37;
  dev->INI_PowerDelayLines = A4CIS_PowerDelayLines;
  dev->INI_HomeLines = A4CIS_HomeLines;
  dev->INI_DarkLines = 50;
  dev->INI_KLevel = A4CIS_KLevel;
  dev->INI_MaxPowerDelay = A4CIS_MaxPowerDelay;
  dev->INI_MinPowerDelay = A4CIS_MaxPowerDelay;
  dev->INI_AdjustWay = 0;
  dev->INI_GreenBlackFactor = 0.0;
  dev->INI_BlueBlackFactor = 0.0;
  dev->INI_RedBlackFactor = 0.0;
  dev->INI_GrayBlackFactor = 0.0;
  dev->INI_GreenFactor = 0.0;
  dev->INI_BlueFactor = 0.0;
  dev->INI_RedFactor = 0.0;
  dev->INI_GrayFactor = 0.0;
  
  dev->INI_RedRgb600PGA = 0;
  dev->INI_GreenRgb600PGA = 0;
  dev->INI_BlueRgb600PGA = 0;
  dev->INI_Mono600PGA = 0;
  dev->INI_RedRgb300PGA = 0;
  dev->INI_GreenRgb300PGA = 0;
  dev->INI_BlueRgb300PGA = 0;
  dev->INI_Mono300PGA = 0;
  dev->INI_ExploseTime = 3200;
  dev->INI_RedRgb600PowerDelay = 0;
  dev->INI_GreenRgb600PowerDelay = 0;
  dev->INI_BlueRgb600PowerDelay = 0;
  dev->INI_RedMono600PowerDelay = 0;
  dev->INI_GreenMono600PowerDelay = 0;
  dev->INI_BlueMono600PowerDelay = 0;
  dev->INI_RedRgb300PowerDelay = 0;
  dev->INI_GreenRgb300PowerDelay = 0;
  dev->INI_BlueRgb300PowerDelay = 0;
  dev->INI_RedMono300PowerDelay = 0;
  dev->INI_GreenMono300PowerDelay = 0;
  dev->INI_BlueMono300PowerDelay = 0;
  dev->INI_Threshold = 128;
  
  dev->INI_TopRef = 0x10;
  dev->INI_FrontEnd = 0x10;
  dev->INI_RedOffset = 0x00;
  dev->INI_GreenOffset = 0x00;
  dev->INI_BlueOffset = 0x00;
  
  dev->INI_Rgb36BitBackTrack = 80;
  dev->INI_Rgb24BitBackTrack = 80;
  dev->INI_Mono8BitBackTrack = 80;
  dev->INI_Mono4BitBackTrack = 80;
  dev->INI_Mono1BitBackTrack = 80;

  dev->isOpened = SANE_FALSE;
  dev->isPrepared = SANE_FALSE;
  dev->ExploseTime = 4000;
  dev->Width = 2550;
  dev->XDpi = 300;
  dev->YDpi = 300;
  dev->ScanMode = CRGB24ext;
  dev->BytesPerRow = 2550 * 3;
  dev->Dummy = 0;
  dev->BytesPerStrip = 2550;
  dev->pImageBuffer = NULL;
  dev->pRed = NULL;
  dev->pGreen = NULL;
  dev->pBlue = NULL;
  dev->pfGetLine = NULL;
  dev->pfBackTrack = NULL;
  dev->isAdjustedRgb600PowerDelay = SANE_FALSE;
  dev->isAdjustedMono600PowerDelay = SANE_FALSE;
  dev->isAdjustedRgb300PowerDelay = SANE_FALSE;
  dev->isAdjustedMono300PowerDelay = SANE_FALSE;
  dev->isEvaluePixelRate = SANE_FALSE;
  dev->RedRgb600PGA = 0;
  dev->GreenRgb600PGA = 0;
  dev->BlueRgb600PGA = 0;
  dev->Mono600PGA = 0;
  dev->RedRgb600PowerDelay = 0;
  dev->GreenRgb600PowerDelay = 0;
  dev->BlueRgb600PowerDelay = 0;
  dev->RedMono600PowerDelay = 0;
  dev->GreenMono600PowerDelay = 0;
  dev->BlueMono600PowerDelay = 0;
  dev->RedRgb300PGA = 0;
  dev->GreenRgb300PGA = 0;
  dev->BlueRgb300PGA = 0;
  dev->Mono300PGA = 0;
  dev->RedRgb300PowerDelay = 0;
  dev->GreenRgb300PowerDelay = 0;
  dev->BlueRgb300PowerDelay = 0;
  dev->RedMono300PowerDelay = 0;
  dev->GreenMono300PowerDelay = 0;
  dev->BlueMono300PowerDelay = 0;
  dev->PixelRate = 2000;
  dev->Threshold = 128;
  dev->pGammaTable = NULL;
  dev->SkipsPerRow = 0;

  dev->pRedCalibrator = NULL;
  dev->pGreenCalibrator = NULL;
  dev->pBlueCalibrator = NULL;
  dev->pMonoCalibrator = NULL;

  dev->Sensor = SR_CANON600;
  dev->bAlreadyDetectCIS = SANE_FALSE;

  RIE(usb_low_init (&dev->chip));

  DBG (5, "usb_high_scan_init: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_exit (Mustek_Usb_Device * dev)
{
  SANE_Status status;
  
  DBG (5, "usb_high_scan_exit: start\n");
  RIE(usb_low_exit (dev->chip));
  dev->chip = 0;
  DBG (5, "usb_high_scan_exit: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_Prepare (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_Prepare: start dev=%p\n", dev);
  if (dev->isPrepared)
    {
      DBG (5,"usb_high_scan_Prepare: is already prepared\n");
      return SANE_STATUS_GOOD;
    }
  if (dev->pImageBuffer)
    {
      free (dev->pImageBuffer);
    }
  dev->pImageBuffer = (SANE_Byte *) malloc (A4CIS_BytesPerStrip * 3);
  if (!dev->pImageBuffer)
    return SANE_STATUS_NO_MEM;

  dev->pRed = dev->pImageBuffer;
  dev->pGreen = dev->pImageBuffer + A4CIS_BytesPerStrip;
  dev->pBlue = dev->pImageBuffer + A4CIS_BytesPerStrip * 2;

  dev->isPrepared = SANE_TRUE;
  DBG (5, "usb_high_scan_Prepare: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_Clearup (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_Clearup: start, dev=%p\n", dev);
  if (!dev->isPrepared)
    {
      DBG (3, "usb_high_scan_Clearup: is not prepared\n");
      return SANE_STATUS_INVAL;
    }
  if (dev->pImageBuffer)
    {
      free (dev->pImageBuffer);
    }
  dev->pImageBuffer = NULL;
  dev->pRed = NULL;
  dev->pGreen = NULL;
  dev->pBlue = NULL;

  dev->isPrepared = SANE_FALSE;
  DBG (5, "usb_high_scan_Clearup: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_TurnPower (Mustek_Usb_Device * dev, SANE_Bool isOn)
{
  SANE_Status status;

  DBG (5, "usb_high_scan_TurnPower: start, turn %s power\n", 
       isOn ? "on" : "off");

  if (dev->isOpened)
    {
      DBG (3, "usb_high_scan_TurnPower: isOpened\n");
      return SANE_STATUS_INVAL;
    }
  RIE(usb_low_Open (dev->chip, dev->DeviceName));

  if (isOn)
    {
      RIE(usb_low_TurnPeripheralPower (dev->chip, SANE_TRUE));
      RIE(usb_low_TurnLampPower (dev->chip, SANE_TRUE));
    }
  else
    {
      RIE(usb_low_TurnLampPower (dev->chip, SANE_FALSE));
    }
  RIE(usb_low_Close (dev->chip));

  DBG (5, "usb_high_scan_TurnPower: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_BackHome (Mustek_Usb_Device * dev)
{
  SANE_Status status;

  DBG (5, "usb_high_scan_BackHome: start\n");

  if (dev->isOpened)
    {
      DBG (3, "usb_high_scan_BackHome: is already open\n");
      return SANE_STATUS_INVAL;
    }
  RIE(usb_low_Open (dev->chip, dev->DeviceName));

  RIE(usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime));
  RIE(usb_mid_motor1200_PrepareHome (dev->chip));
  RIE(usb_low_Close (dev->chip));

  DBG (5, "usb_high_scan_BackHome: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_SetThreshold (Mustek_Usb_Device * dev, SANE_Byte threshold)
{
  DBG (5, "usb_high_scan_SetThreshold: start, dev=%p, threshold=%d\n",
       dev, threshold);

  if (dev->isOpened)
    {
      DBG (3, "usb_high_scan_SetThreshold: is already opened\n");
      return SANE_STATUS_INVAL;
    }

  dev->Threshold = threshold;
  DBG (5, "usb_high_scan_SetThreshold: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_EmbedGamma (Mustek_Usb_Device * dev, SANE_Word * pGammaTable)
{
  DBG (5, "usb_high_scan_EmbedGamma: start, dev=%p, pGammaTable=%d\n",
       dev, pGammaTable);
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_EmbedGamma isOpened\n");
      return SANE_STATUS_INVAL;
    }
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_EmbedGamma !isPrepared\n");
      return SANE_STATUS_INVAL;
    }

  dev->pGammaTable = pGammaTable;
  DBG (5, "usb_high_scan_EmbedGamma: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_Reset (Mustek_Usb_Device * dev)
{
  SANE_Status status;

  DBG (5, "usb_high_scan_Reset: start\n");

  if (dev->isOpened)
    {
      DBG (3, "usb_high_scan_Reset: isOpened\n");
      return SANE_STATUS_INVAL;
    }
  if (!dev->isPrepared)
    {
      DBG (3, "usb_high_scan_Reset: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }
  RIE(usb_low_Open (dev->chip, dev->DeviceName));
  RIE(usb_high_scan_InitASIC (dev, dev->Sensor));
  RIE(usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime));
  RIE(usb_mid_motor1200_PrepareHome (dev->chip));
  RIE(usb_low_Close (dev->chip));
  RIE(usb_high_scan_SetThreshold (dev, dev->INI_Threshold));
  RIE(usb_high_scan_EmbedGamma (dev, NULL));
  dev->isAdjustedRgb600PowerDelay = SANE_FALSE;
  dev->isAdjustedMono600PowerDelay = SANE_FALSE;
  dev->isAdjustedRgb300PowerDelay = SANE_FALSE;
  dev->isAdjustedMono300PowerDelay = SANE_FALSE;
  dev->isEvaluePixelRate = SANE_FALSE;
  DBG (5, "usb_high_scan_Reset: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_WaitCarriageHome (Mustek_Usb_Device * dev)
{
  SANE_Status status;
  
  DBG (5, "usb_high_scan_WaitCarriageHome: start\n");

  status = usb_low_GetHomeSensor (dev->chip);

  if (status != SANE_STATUS_GOOD)
    {
      RIE(usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime));
      RIE(usb_mid_motor1200_PrepareHome (dev->chip));
      do
	{
	  status = usb_low_GetHomeSensor (dev->chip);
	  if (status != SANE_STATUS_GOOD)
	    usleep (18 * 1000);
	}
      while (status != SANE_STATUS_GOOD);
    }

  /* No Motor & Forward */
  RIE(usb_low_MoveMotorHome (dev->chip, SANE_FALSE, SANE_FALSE));
  DBG (5, "usb_high_scan_WaitCarriageHome: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_HardwareCalibration (Mustek_Usb_Device * dev)
{
  SANE_Status status;
  
  DBG (5, "usb_high_scan_HardwareCalibration: start\n");

  RIE(usb_high_scan_SafeForward (dev, dev->INI_HomeLines));

  switch (dev->INI_AdjustWay)
    {
    case 1:
      switch (dev->ScanMode)
	{
	case CRGB24ext:
	  if (dev->Sensor == SR_CANON300600)
	    {
	      if (usb_mid_c300600_is600Mode (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb600PGA = dev->INI_RedRgb600PGA;
		  dev->GreenRgb600PGA = dev->INI_GreenRgb600PGA;
		  dev->BlueRgb600PGA = dev->INI_BlueRgb600PGA;
		  RIE(usb_high_scan_AdjustRgb600PowerDelay (dev));
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb300PGA = dev->INI_RedRgb300PGA;
		  dev->GreenRgb300PGA = dev->INI_GreenRgb300PGA;
		  dev->BlueRgb300PGA = dev->INI_BlueRgb300PGA;
		  RIE(usb_high_scan_AdjustRgb300PowerDelay (dev));
		}
	    }
	  else if (dev->Sensor == SR_CANON300)
	    {
	      if (usb_mid_c300_is600Mode (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb600PGA = dev->INI_RedRgb600PGA;
		  dev->GreenRgb600PGA = dev->INI_GreenRgb600PGA;
		  dev->BlueRgb600PGA = dev->INI_BlueRgb600PGA;
		  RIE(usb_high_scan_AdjustRgb600PowerDelay (dev));
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb300PGA = dev->INI_RedRgb300PGA;
		  dev->GreenRgb300PGA = dev->INI_GreenRgb300PGA;
		  dev->BlueRgb300PGA = dev->INI_BlueRgb300PGA;
		  RIE(usb_high_scan_AdjustRgb300PowerDelay (dev));
		}
	    }
	  else
	    {
	      if (usb_mid_c600_is600Mode (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb600PGA = dev->INI_RedRgb600PGA;
		  dev->GreenRgb600PGA = dev->INI_GreenRgb600PGA;
		  dev->BlueRgb600PGA = dev->INI_BlueRgb600PGA;
		  RIE(usb_high_scan_AdjustRgb600PowerDelay (dev));
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->RedRgb300PGA = dev->INI_RedRgb300PGA;
		  dev->GreenRgb300PGA = dev->INI_GreenRgb300PGA;
		  dev->BlueRgb300PGA = dev->INI_BlueRgb300PGA;
		  RIE(usb_high_scan_AdjustRgb300PowerDelay (dev));
		}
	    }
	  break;
	case CGRAY8ext:
	  if (dev->Sensor == SR_CANON300600)
	    {
	      if (usb_mid_c300600_is600Mode (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono600PGA = dev->INI_Mono600PGA;
		  RIE(usb_high_scan_EvaluePixelRate (dev));
		  RIE(usb_high_scan_AdjustMono600PowerDelay (dev));
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono300PGA = dev->INI_Mono300PGA;
		  RIE(usb_high_scan_EvaluePixelRate (dev));
		  RIE(usb_high_scan_AdjustMono300PowerDelay (dev));
		}
	    }
	  else if (dev->Sensor == SR_CANON300)
	    {
	      if (usb_mid_c300_is600Mode (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono600PGA = dev->INI_Mono600PGA;
		  RIE(usb_high_scan_EvaluePixelRate (dev));
		  RIE(usb_high_scan_AdjustMono600PowerDelay (dev));
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono300PGA = dev->INI_Mono300PGA;
		  RIE(usb_high_scan_EvaluePixelRate (dev));
		  RIE(usb_high_scan_AdjustMono300PowerDelay (dev));
		}
	    }
	  else
	    {
	      if (usb_mid_c600_is600Mode
		  (dev->chip, dev->XDpi))
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono600PGA = dev->INI_Mono600PGA;
		  RIE(usb_high_scan_EvaluePixelRate (dev));
		  RIE(usb_high_scan_AdjustMono600PowerDelay (dev));
		}
	      else
		{
		  dev->ExploseTime = dev->INI_ExploseTime;
		  dev->Mono300PGA = dev->INI_Mono300PGA;
		  RIE(usb_high_scan_EvaluePixelRate (dev));
		  RIE(usb_high_scan_AdjustMono300PowerDelay (dev));
		}
	    }
	  break;
	default:
	  break;
	}
      break;
    default:
      dev->ExploseTime = dev->INI_ExploseTime;
      dev->RedRgb600PowerDelay = dev->INI_RedRgb600PowerDelay;
      dev->GreenRgb600PowerDelay = dev->INI_GreenRgb600PowerDelay;
      dev->BlueRgb600PowerDelay = dev->INI_BlueRgb600PowerDelay;
      dev->RedMono600PowerDelay = dev->INI_RedMono600PowerDelay;
      dev->GreenMono600PowerDelay = dev->INI_GreenMono600PowerDelay;
      dev->BlueMono600PowerDelay = dev->INI_BlueMono600PowerDelay;
      dev->RedRgb600PGA = dev->INI_RedRgb600PGA;
      dev->GreenRgb600PGA = dev->INI_GreenRgb600PGA;
      dev->BlueRgb600PGA = dev->INI_BlueRgb600PGA;
      dev->Mono600PGA = dev->INI_Mono600PGA;
      dev->RedRgb300PowerDelay = dev->INI_RedRgb300PowerDelay;
      dev->GreenRgb300PowerDelay = dev->INI_GreenRgb300PowerDelay;
      dev->BlueRgb300PowerDelay = dev->INI_BlueRgb300PowerDelay;
      dev->RedMono300PowerDelay = dev->INI_RedMono300PowerDelay;
      dev->GreenMono300PowerDelay = dev->INI_GreenMono300PowerDelay;
      dev->BlueMono300PowerDelay = dev->INI_BlueMono300PowerDelay;
      dev->RedRgb300PGA = dev->INI_RedRgb300PGA;
      dev->GreenRgb300PGA = dev->INI_GreenRgb300PGA;
      dev->BlueRgb300PGA = dev->INI_BlueRgb300PGA;
      dev->Mono300PGA = dev->INI_Mono300PGA;
      break;
    }
  DBG (5, "usb_high_scan_HardwareCalibration: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_LineCalibration (Mustek_Usb_Device * dev)
{
  SANE_Status status;
  
  DBG (5, "usb_high_scan_LineCalibration: start\n");
  switch (dev->ScanMode)
    {
    case CRGB24ext:
      RIE(usb_high_scan_PrepareRgb24Bit (dev));
      if (dev->Sensor == SR_CANON300600)
	{
	  if (usb_mid_c300600_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareRgbSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareRgbSignal300Dpi (dev));
	  RIE(usb_mid_c300600_PrepareRgbX (dev->chip, dev->XDpi));
	}
      else if (dev->Sensor == SR_CANON300)
	{
	  if (usb_mid_c300_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareRgbSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareRgbSignal300Dpi (dev));
	  RIE(usb_mid_c300_PrepareRgbX (dev->chip, dev->XDpi));
	}
      else
	{
	  if (usb_mid_c600_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareRgbSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareRgbSignal300Dpi (dev));
	  RIE( usb_mid_c600_PrepareRgbX (dev->chip, dev->XDpi));
	}
      RIE(usb_high_scan_CalibrationRgb24Bit (dev));
      break;
    case CGRAY8ext:
      RIE(usb_high_scan_PrepareMono8Bit (dev));
      if (dev->Sensor == SR_CANON300600)
	{
	  if (usb_mid_c300600_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareMonoSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareMonoSignal300Dpi (dev));
	  RIE(usb_mid_c300600_PrepareMonoX (dev->chip, dev->XDpi));
	}
      else if (dev->Sensor == SR_CANON300)
	{
	  if (usb_mid_c300_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareMonoSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareMonoSignal300Dpi (dev));
	  RIE(usb_mid_c300_PrepareMonoX (dev->chip, dev->XDpi));
	}
      else
	{
	  if (usb_mid_c600_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareMonoSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareMonoSignal300Dpi (dev));
	  RIE(usb_mid_c600_PrepareMonoX (dev->chip, dev->XDpi));
	}
      RIE(usb_high_scan_CalibrationMono8Bit (dev));
      break;
    default:
      DBG (3, "usb_high_scan_LineCalibration: mode not matched\n");
      return SANE_STATUS_INVAL;
      break;
    }
  DBG (5, "usb_high_scan_LineCalibration: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_PrepareScan (Mustek_Usb_Device * dev)
{
  SANE_Status status;
 
  DBG (5, "usb_high_scan_PrepareScan: start\n");
  switch (dev->ScanMode)
    {
    case CRGB24ext:
      RIE(usb_high_scan_PrepareRgb24Bit (dev));
      dev->pfGetLine = &usb_high_scan_GetRgb24BitLine;
      dev->pfBackTrack = &usb_high_scan_BackTrackRgb24Bit;

      if (dev->Sensor == SR_CANON300600)
	{
	  if (usb_mid_c300600_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareRgbSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareRgbSignal300Dpi (dev));
	  RIE(usb_mid_c300600_PrepareRgbX (dev->chip, dev->XDpi));
	}
      else if (dev->Sensor == SR_CANON300)
	{
	  if (usb_mid_c300_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareRgbSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareRgbSignal300Dpi (dev));
	  RIE(usb_mid_c300_PrepareRgbX (dev->chip, dev->XDpi));
	}
      else
	{
	  if (usb_mid_c600_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareRgbSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareRgbSignal300Dpi (dev));
	  RIE(usb_mid_c600_PrepareRgbX (dev->chip, dev->XDpi));
	}
      if (dev->chip->scanner_type == mustek_usb_type_600cu)
	RIE(usb_mid_motor600_PrepareRgbY (dev->chip, dev->YDpi));
      else
	RIE(usb_mid_motor1200_PrepareRgbY (dev->chip, dev->YDpi));
      break;
    case CGRAY8ext:
      RIE(usb_high_scan_PrepareMono8Bit (dev));
      dev->pfGetLine = &usb_high_scan_GetMono8BitLine;
      dev->pfBackTrack = &usb_high_scan_BackTrackMono8Bit;
      if (dev->Sensor == SR_CANON300600)
	{
	  if (usb_mid_c300600_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareMonoSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareMonoSignal300Dpi (dev));
	  RIE(usb_mid_c300600_PrepareMonoX (dev->chip, dev->XDpi));
	}
      else if (dev->Sensor == SR_CANON300)
	{
	  if (usb_mid_c300_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareMonoSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareMonoSignal300Dpi (dev));
	  RIE(usb_mid_c300600_PrepareMonoX (dev->chip, dev->XDpi));
	}
      else
	{
	  if (usb_mid_c600_is600Mode (dev->chip, dev->XDpi))
	    RIE(usb_high_scan_PrepareMonoSignal600Dpi (dev));
	  else
	    RIE(usb_high_scan_PrepareMonoSignal300Dpi (dev));
	  RIE(usb_mid_c600_PrepareMonoX (dev->chip, dev->XDpi));
	}
      if (dev->chip->scanner_type == mustek_usb_type_600cu)
	RIE(usb_mid_motor600_PrepareMonoY (dev->chip, dev->YDpi));
      else
	RIE(usb_mid_motor1200_PrepareMonoY (dev->chip, dev->YDpi));
      break;
    default:
      DBG (5, "usb_high_scan_PrepareScan: unmatched mode\n");
      return SANE_STATUS_INVAL;
      break;
    }
  DBG (5, "usb_high_scan_PrepareScan: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_ScanSuggest (Mustek_Usb_Device * dev, PTARGETIMAGE pTarget,
			   PSUGGESTSETTING pSuggest)
{
  int i;
  DBG (5, "usb_high_scan_ScanSuggest: start\n");

  /* Looking up Optical Y Resolution */
  for (i = 0; usb_mid_motor1200_s_wOpticalYDpi[i] != 0; i++)
    {
      if (usb_mid_motor1200_s_wOpticalYDpi[i] <= pTarget->wDpi)
	{
	  pSuggest->wYDpi = usb_mid_motor1200_s_wOpticalYDpi[i];
	  break;
	}
    }
  if (usb_mid_motor1200_s_wOpticalYDpi[i] == 0)
    {
      i--;
      pSuggest->wYDpi = usb_mid_motor1200_s_wOpticalYDpi[i];
    }

  /* Looking up Optical X Resolution */

  if (dev->Sensor == SR_CANON300600)
    {
      for (i = 0; usb_mid_c300600_s_wOpticalXDpi[i] != 0; i++)
	{
	  if (usb_mid_c300600_s_wOpticalXDpi[i] <= pTarget->wDpi)
	    {
	      pSuggest->wXDpi = usb_mid_c300600_s_wOpticalXDpi[i];
	      break;
	    }
	}
      if (usb_mid_c300600_s_wOpticalXDpi[i] == 0)
	{
	  i--;
	  pSuggest->wXDpi = usb_mid_c300600_s_wOpticalXDpi[i];
	}
    }
  else if (dev->Sensor == SR_CANON300)
    {
      for (i = 0; usb_mid_c300_s_wOpticalXDpi[i] != 0; i++)
	{
	  if (usb_mid_c300_s_wOpticalXDpi[i] <= pTarget->wDpi)
	    {
	      pSuggest->wXDpi = usb_mid_c300_s_wOpticalXDpi[i];
	      break;
	    }
	}
      if (usb_mid_c300_s_wOpticalXDpi[i] == 0)
	{
	  i--;
	  pSuggest->wXDpi = usb_mid_c300_s_wOpticalXDpi[i];
	}
    }
  else
    {
      for (i = 0; usb_mid_c600_s_wOpticalXDpi[i] != 0; i++)
	{
	  if (usb_mid_c600_s_wOpticalXDpi[i] <= pTarget->wDpi)
	    {
	      pSuggest->wXDpi = usb_mid_c600_s_wOpticalXDpi[i];
	      break;
	    }
	}
      if (usb_mid_c600_s_wOpticalXDpi[i] == 0)
	{
	  i--;
	  pSuggest->wXDpi = usb_mid_c600_s_wOpticalXDpi[i];
	}
    }
  pSuggest->wX = (SANE_Word) (((SANE_Word) (pTarget->wX) 
			       * (SANE_Word) (pSuggest->wXDpi)) 
			      / (SANE_Word) (pTarget->wDpi));
  pSuggest->wY = (SANE_Word) (((SANE_Word) (pTarget->wY) 
			       * (SANE_Word) (pSuggest->wYDpi)) 
			      / (SANE_Word) (pTarget->wDpi));
  pSuggest->wWidth = (SANE_Word) (((SANE_Word) (pTarget->wWidth) 
				   * (SANE_Word) (pSuggest->wXDpi)) 
				  / (SANE_Word) (pTarget->wDpi));
  pSuggest->wHeight = (SANE_Word) (((SANE_Word) (pTarget->wHeight) 
				    * (SANE_Word) (pSuggest->wYDpi)) 
				   / (SANE_Word) (pTarget->wDpi));

  if (pTarget->isOptimalSpeed)
    {
      switch (pTarget->cmColorMode)
	{
	case CRGB24:
	  pSuggest->cmScanMode = CRGB24ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth) * 3;
	  break;
	case CGRAY8:
	  pSuggest->cmScanMode = CGRAY8ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth);
	  break;
	default:
	  DBG (3, "usb_high_scan_ScanSuggest: unmatched mode\n");
	  return SANE_STATUS_INVAL;
	  break;
	}
    }
  else
    {
      switch (pTarget->cmColorMode)
	{
	case CRGB24:
	  pSuggest->cmScanMode = CRGB24ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth) * 3;
	  break;
	case CGRAY8:
	  pSuggest->cmScanMode = CGRAY8ext;
	  pSuggest->dwBytesPerRow = (SANE_Word) (pSuggest->wWidth);
	  break;
	default:
	  DBG (3, "usb_high_scan_ScanSuggest: unmatched mode\n");
	  return SANE_STATUS_INVAL;
	  break;
	}
    }
  DBG (5, "usb_high_scan_ScanSuggest: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_detect_sensor (Mustek_Usb_Device *dev)
{
  SANE_Byte *pBuffer;
  static SANE_Word L_Temp = 0, R_Temp = 0;
  int i;
  SANE_Status status;
  SANE_Word lines_left;

  DBG (5, "usb_high_scan_detect_sensor: start\n");

  if (dev->chip->scanner_type == mustek_usb_type_600cu)
    {
      DBG (5, "usb_high_scan_detect_sensor: Sensor looks like Canon "
	   "300 dpi\n");
      dev->Sensor = SR_CANON300;
      return SANE_STATUS_GOOD;
    }

  pBuffer = NULL;
  L_Temp = 0;
  R_Temp = 0;

  if (dev->bAlreadyDetectCIS)
    {
      DBG (5, "usb_high_scan_detect_sensor: sensor already detected\n");
      return SANE_STATUS_GOOD;
    }

  pBuffer = (SANE_Byte *) malloc (A4CIS_BytesPerStrip);

  if (!pBuffer)
    return SANE_STATUS_NO_MEM;
  
  for (i = 0; i < 5400; i++)
    pBuffer[i] = 0xaa;
  
  dev->ScanMode = CGRAY8ext;
  dev->XDpi = 600;
  dev->YDpi = 1200;
  dev->Width = 5400;
  RIE(usb_low_Open (dev->chip, dev->DeviceName));

  dev->isOpened = SANE_TRUE;
  RIE(usb_high_scan_InitASIC (dev, SR_CANON600));
  RIE(usb_low_TurnPeripheralPower (dev->chip, SANE_TRUE));
  RIE(usb_low_EnableMotor (dev->chip, SANE_TRUE));	/* Enable Motor */
  RIE(usb_low_TurnLampPower (dev->chip, SANE_TRUE));
  RIE(usb_low_InvertImage (dev->chip, SANE_FALSE));
  RIE(usb_low_SetImageDpi (dev->chip, SANE_TRUE, SW_P6P6));
  dev->BytesPerStrip = A4CIS_600AdjustLength;
  dev->BytesPerRow = 5400;
  dev->Dummy = 0;
  
  RIE(usb_high_scan_WaitCarriageHome (dev));
  RIE(usb_high_scan_HardwareCalibration (dev));
  RIE(usb_high_scan_PrepareScan (dev));
  
  /* Get Data */
  RIE(usb_low_StartRowing (dev->chip));
  RIE(usb_low_GetRow (dev->chip, pBuffer, &lines_left));
  RIE(usb_low_StopRowing (dev->chip));
  /* Calculate */
  for (i = 0; i < 256; i++)
    L_Temp = L_Temp + pBuffer[512 + i];
  for (i = 0; i < 256; i++)
    R_Temp = R_Temp + pBuffer[3500 + i];
  
  L_Temp = L_Temp / 256;
  R_Temp = R_Temp / 256;
  
  /* 300/600 switch CIS or 600 CIS */
  DBG (5, "usb_high_scan_detect_sensor: L_Temp=%d, R_Temp=%d\n",
       L_Temp, R_Temp);
  if (R_Temp > 50)
    {
      dev->Sensor = SR_CANON600;
      DBG (5, "usb_high_scan_detect_sensor: Sensor looks like Canon 600 "
	   "dpi\n");
    }
  else
    {
      DBG (5, "usb_high_scan_detect_sensor: Sensor looks like Canon "
	   "300/600 dpi\n");
      dev->Sensor = SR_CANON300600;
    }
  /* Release Resource */
  usb_low_Close (dev->chip);
  free (pBuffer);
  pBuffer = NULL;
  
  usb_low_SetCIS (dev->chip, dev->Sensor);
  dev->bAlreadyDetectCIS = SANE_TRUE;
  DBG (5, "usb_high_scan_detect_sensor: exit\n");
  return SANE_STATUS_GOOD;
}


SANE_Status
usb_high_scan_SetupScan (Mustek_Usb_Device * dev, COLORMODE ColorMode,
			 SANE_Word XDpi, SANE_Word YDpi, SANE_Bool isInvert,
			 SANE_Word X, SANE_Word Y, SANE_Word Width)
{
  SANE_Status status;
  SANE_Word UpperBound;
  SANE_Word LeftBound;

  DBG (5, "usb_high_scan_SetupScan: start, isInvert=%d\n", isInvert);
  if (dev->isOpened)
    {
      DBG (5,"usb_high_scan_SetupScan: isOpened\n");
      return SANE_STATUS_INVAL;
    }
  if (!dev->isPrepared)
    {
      DBG (5,"usb_high_scan_SetupScan: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }

  RIE(usb_high_scan_detect_sensor (dev));
  RIE(usb_low_Open (dev->chip, dev->DeviceName));
  dev->isOpened = SANE_TRUE;

  RIE(usb_high_scan_InitASIC (dev, dev->Sensor));
  RIE(usb_low_TurnPeripheralPower (dev->chip, SANE_TRUE));
  RIE(usb_low_EnableMotor (dev->chip, SANE_TRUE));	/* Enable Motor */
  RIE(usb_low_TurnLampPower (dev->chip, SANE_TRUE));
  RIE(usb_low_InvertImage (dev->chip, SANE_FALSE));

  dev->ScanMode = ColorMode;
  dev->XDpi = XDpi;
  dev->YDpi = YDpi;
  dev->Width = Width;

  /* Compute necessary variable */
  UpperBound = ((SANE_Word) (((SANE_Word) (Y) * 600L) 
			    / (SANE_Word) (dev->YDpi)) 
		+ dev->INI_JLines);	/* in 600dpi */

  switch (dev->ScanMode)
    {
    case CRGB24ext:
      dev->BytesPerRow = 3 * dev->Width;
      break;
    case CGRAY8ext:
      dev->BytesPerRow = dev->Width;
      break;
    default:
      break;
    }

  if (dev->Sensor == SR_CANON300600)
    {
      if (usb_mid_c300600_is600Mode (dev->chip, dev->XDpi))
	{
	  LeftBound =
	    (SANE_Word) (((SANE_Word) (X) * 600L) / (SANE_Word) (dev->XDpi)) 
	    + dev->INI_600SkipsPerRow;	/* in 600dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (((SANE_Word) (LeftBound % 32) 
			  * (SANE_Word) (dev->XDpi) + 300L) / 600L);
	}
      else
	{
	  LeftBound =
	    (SANE_Word) (((SANE_Word) (X) * 300L) / (SANE_Word) (dev->XDpi)) +
	    dev->INI_300SkipsPerRow;	/* in 300dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (((SANE_Word) (LeftBound % 32) 
			  * (SANE_Word) (dev->XDpi) +  150L) / 300L);
	}
    }
  else if (dev->Sensor == SR_CANON300)
    {
      if (usb_mid_c300_is600Mode (dev->chip, dev->XDpi))
	{
	  LeftBound =
	    (SANE_Word) (((SANE_Word) (X) * 600L) / (SANE_Word) (dev->XDpi)) 
	    + dev->INI_600SkipsPerRow;	/* in 600dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (((SANE_Word) (LeftBound % 32) 
			  * (SANE_Word) (dev->XDpi) + 300L) / 600L);
	}
      else
	{
	  LeftBound =
	    (SANE_Word) (((SANE_Word) (X) * 300L) / (SANE_Word) (dev->XDpi)) +
	    dev->INI_300SkipsPerRow;	/* in 300dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (((SANE_Word) (LeftBound % 32) 
			  * (SANE_Word) (dev->XDpi) +  150L) / 300L);
	}
    }
  else
    {
      if (usb_mid_c600_is600Mode (dev->chip, dev->XDpi))
	{
	  LeftBound =
	    (SANE_Word) (((SANE_Word) (X) * 600L) / (SANE_Word) (dev->XDpi))
	    + dev->INI_600SkipsPerRow;	/* in 600dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (((SANE_Word) (LeftBound % 32) 
			  * (SANE_Word) (dev->XDpi) + 300L) / 600L);
	}
      else
	{
	  LeftBound = 
	    (SANE_Word) (((SANE_Word) X) * 300L / (SANE_Word) dev->XDpi) +
	    dev->INI_300SkipsPerRow;	/* in 300dpi */
	  dev->SkipsPerRow =
	    (SANE_Word) (((SANE_Word) LeftBound % 32 
			  * ((SANE_Word) dev->XDpi) + 150L) / 300L);
	}
    }
  
  dev->Dummy = (LeftBound / 32) * 32;

  switch (dev->ScanMode)
    {
    case CRGB24ext:
      dev->BytesPerStrip = dev->SkipsPerRow + dev->Width;
      break;
    case CGRAY8ext:
      dev->BytesPerStrip = dev->SkipsPerRow + dev->Width;
      break;
    default:
      break;
    }

  dev->BytesPerStrip = ((dev->BytesPerStrip + 1) / 2) * 2;
  /* make BytesPerStrip is as 2n to advoid 64n+1 */

  RIE(usb_high_scan_WaitCarriageHome (dev));
  RIE(usb_high_scan_HardwareCalibration (dev));
  RIE(usb_high_scan_LineCalibration (dev));
  RIE(usb_high_scan_StepForward (dev, UpperBound));
  RIE(usb_high_scan_PrepareScan (dev));
  RIE(usb_low_StartRowing (dev->chip));
  DBG (5, "usb_high_scan_SetupScan: exit\n");
  return SANE_STATUS_GOOD;
}
  
SANE_Status
usb_high_scan_GetRows (Mustek_Usb_Device * dev, SANE_Byte * lpBlock,
		       SANE_Word Rows, SANE_Bool isOrderInvert)
{
  SANE_Status status;

  DBG (5, "usb_high_scan_GetRows: start, %d rows\n", Rows);
  if (!dev->isOpened)
    {
      DBG (3, "usb_high_scan_GetRows: !isOpened\n");
      return SANE_STATUS_INVAL;
    }
  if (!dev->isPrepared)
    {
      DBG (3, "usb_high_scan_GetRows: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }
  while (Rows > 0)
    {
      RIE((*dev->pfGetLine) (dev, lpBlock, isOrderInvert));
      lpBlock += dev->BytesPerRow;
      Rows--;
    }
  DBG (5, "usb_high_scan_GetRows: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_StopScan (Mustek_Usb_Device * dev)
{
  SANE_Status status;
  
  DBG (5, "usb_high_scan_StopScan: start\n");
  if (!dev->isOpened)
    {
      DBG (3, "usb_high_scan_StopScan: !isOpened\n");
      return SANE_STATUS_INVAL;
    }
  if (!dev->isPrepared)
    {
      DBG (3, "usb_high_scan_StopScan: !isPrepared\n");
      return SANE_STATUS_INVAL;
    }
  switch (dev->ScanMode)
    {
    case CRGB24ext:
      RIE(usb_high_cal_MD12_exit (dev->pBlueCalibrator));
      if (dev->pBlueCalibrator)
	free (dev->pBlueCalibrator);
      dev->pBlueCalibrator = NULL;
      RIE(usb_high_cal_MD12_exit (dev->pGreenCalibrator));
      if (dev->pGreenCalibrator)
	free (dev->pGreenCalibrator);
      dev->pGreenCalibrator = NULL;
      RIE(usb_high_cal_MD12_exit (dev->pRedCalibrator));
      if (dev->pRedCalibrator)
	free (dev->pRedCalibrator);
      dev->pRedCalibrator = NULL;
      break;
    case CGRAY8ext:
      RIE(usb_high_cal_MD12_exit (dev->pMonoCalibrator));
      if (dev->pMonoCalibrator)
	free (dev->pMonoCalibrator);
      dev->pMonoCalibrator = NULL;
      break;
    default:
      break;
    }

  RIE(usb_low_StopRowing (dev->chip));
  RIE(usb_low_TurnLampPower (dev->chip, SANE_FALSE));
  RIE(usb_low_Close (dev->chip));
  dev->isOpened = SANE_FALSE;

  DBG (5, "usb_high_scan_StopScan: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_LoadPrivateProfile (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_LoadPrivateProfile: start, dev=%p\n", dev);

  dev->INI_MinExploseTime = 4992;
  dev->INI_600SkipsPerRow = 72;
  dev->INI_300SkipsPerRow = 56;
  dev->INI_JLines = 154;
  dev->INI_KLines = 16;
  dev->INI_KFilter = 8;
  dev->INI_KLoops = 2;
  dev->INI_PixelRateLines = 50;
  dev->INI_PixelRateFilts = 37;
  dev->INI_PowerDelayLines = 2;
  dev->INI_HomeLines = 160;
  dev->INI_DarkLines = 50;
  dev->INI_KLevel = 245;
  dev->INI_MaxPowerDelay = 240;
  dev->INI_MinPowerDelay = 136;
  dev->INI_AdjustWay = 1;
  dev->INI_GreenBlackFactor = -2.0;
			    /*-1.48;*/
  dev->INI_BlueBlackFactor = 0.0;
			  /*-1.69;*/
  dev->INI_RedBlackFactor = 0.0;	/*0.7; */
  dev->INI_GrayBlackFactor = 0.0;
			  /*-1.84;*/

  /* For TestAP... */
  /*dev->INI_GreenFactor=0.7556832608;
     dev->INI_BlueFactor=0.7527943848;
     dev->INI_RedFactor=0.754877035;
     dev->INI_GrayFactor=0.756737835; */
  /* For Release... */
  dev->INI_GreenFactor = 0.82004;
  dev->INI_BlueFactor = 0.84954;
  dev->INI_RedFactor = 0.826375;
  dev->INI_GrayFactor = 0.833375;


  dev->INI_RedRgb600PGA = 8;
  dev->INI_GreenRgb600PGA = 8;
  dev->INI_BlueRgb600PGA = 8;
  dev->INI_Mono600PGA = 8;
  dev->INI_RedRgb300PGA = 8;
  dev->INI_GreenRgb300PGA = 8;
  dev->INI_BlueRgb300PGA = 8;
  dev->INI_Mono300PGA = 8;

  dev->INI_ExploseTime = 9024;
  dev->INI_RedRgb600PowerDelay = 80;
  dev->INI_GreenRgb600PowerDelay = 80;
  dev->INI_BlueRgb600PowerDelay = 80;
  dev->INI_RedMono600PowerDelay = 80;
  dev->INI_GreenMono600PowerDelay = 80;
  dev->INI_BlueMono600PowerDelay = 80;
  dev->INI_RedRgb300PowerDelay = 80;
  dev->INI_GreenRgb300PowerDelay = 80;
  dev->INI_BlueRgb300PowerDelay = 80;
  dev->INI_RedMono300PowerDelay = 80;
  dev->INI_GreenMono300PowerDelay = 80;
  dev->INI_BlueMono300PowerDelay = 80;
  dev->INI_Threshold = 128;

  dev->INI_TopRef = 128;
  dev->INI_FrontEnd = 16;
  dev->INI_RedOffset = 0;
  dev->INI_GreenOffset = 0;
  dev->INI_BlueOffset = 0;

  dev->INI_Rgb36BitBackTrack = 80;
  dev->INI_Rgb24BitBackTrack = 80;
  dev->INI_Mono8BitBackTrack = 80;
  dev->INI_Mono4BitBackTrack = 80;
  dev->INI_Mono1BitBackTrack = 80;
  DBG (5, "usb_high_scan_LoadPrivateProfile: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_InitASIC (Mustek_Usb_Device * dev, SENSOR Sensor)
{
  SANE_Byte CcdDpi = 0;
  SANE_Byte Select = 0;
  SANE_Byte Adjust = 0;
  SANE_Byte Pin = 0;
  SANE_Byte Motor = 0;
  SANE_Bool FixPattern = SANE_FALSE;
  SANE_Byte ADTiming = 0;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_InitASIC: start\n");
  switch (Sensor)
    {
    case SR_TOSHIBA600:
      CcdDpi = 32;
      Select = 240;
      Adjust = 0;
      Pin = 18;
      Motor = 0;
      FixPattern = SANE_FALSE;
      ADTiming = 0;
      DBG (5, "usb_high_scan_InitASIC: sensor is set to TOSHIBA600\n");
      break;
    case SR_CANON300:
      CcdDpi = 232;
      Select = 232;
      Adjust = 0;
      Pin = 18;
      Motor = 0;
      FixPattern = SANE_FALSE;
      ADTiming = 1;
      DBG (5, "usb_high_scan_InitASIC: sensor is set to CANON300\n");
      break;
    case SR_CANON300600:
      CcdDpi = 232;
      Select = 232;
      Adjust = 64;
      Pin = 18;
      Motor = 0;
      FixPattern = SANE_FALSE;
      ADTiming = 1;
      DBG (5, "usb_high_scan_InitASIC: sensor is set to CANON300600\n");
      break;
    case SR_CANON600:
      CcdDpi = 232;
      Select = 232;
      Adjust = 64;
      Pin = 18;
      Motor = 0;
      FixPattern = SANE_FALSE;
      ADTiming = 1;
      DBG (5, "usb_high_scan_InitASIC: sensor is set to CANON600\n");
      break;
    default:
      DBG (5, "usb_high_scan_InitASIC: unknown sensor\n");
      return SANE_STATUS_INVAL;
      break;
    }
  RIE(usb_low_AdjustTiming (dev->chip, Adjust));
  RIE(usb_low_SelectTiming (dev->chip, Select));
  RIE(usb_low_SetTiming (dev->chip, CcdDpi));
  RIE(usb_low_SetSRAMBank (dev->chip, BS_16K));
  RIE(usb_low_SetASICIoPins (dev->chip, Pin));
  RIE(usb_low_SetRGBSelPins (dev->chip, Pin));
  RIE(usb_low_SetMotorSignal (dev->chip, Motor));
  RIE(usb_low_SetTestSRAMMode (dev->chip, SANE_FALSE));
  RIE(usb_low_SetFixPattern (dev->chip, FixPattern));
  RIE(usb_low_SetADTiming (dev->chip, ADTiming));

  DBG (5, "usb_high_scan_InitASIC: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_EvalueMaxLevel (Mustek_Usb_Device * dev, SANE_Word SampleLines,
			      int SampleLength, SANE_Byte *ret_max_level)
{
  SANE_Byte MaxLevel = 0;
  SANE_Word i;
  int j;
  SANE_Status status;
  SANE_Word lines_left;

  DBG (5, "usb_high_scan_EvalueMaxLevel: start\n");

  SampleLength -= 20;
  RIE(usb_low_StartRowing (dev->chip));
  for (i = 0; i < SampleLines; i++)
    {
      RIE(usb_low_GetRow (dev->chip, dev->pGreen, &lines_left));
      for (j = 20; j < SampleLength; j++)
	{
	  if (MaxLevel < dev->pGreen[j])
	    MaxLevel = dev->pGreen[j];
	}
    }
  RIE(usb_low_StopRowing (dev->chip));
  if (ret_max_level)
    *ret_max_level = MaxLevel;
  DBG (5, "usb_high_scan_EvalueMaxLevel: exit, MaxLevel = %d\n", MaxLevel);
  return SANE_STATUS_GOOD;
}

/* Binary Search for Single Channel Power Delay */
SANE_Status
usb_high_scan_BSSCPowerDelay (Mustek_Usb_Device * dev,
			      usb_high_PPOWERDELAYFUNC pSetPowerDelay,
			      SIGNALSTATE * ssCurrent, SANE_Byte * Target,
			      SANE_Byte Max, SANE_Byte Min, 
			      SANE_Byte Threshold, int Length)
{
  SANE_Byte MaxLevel;
  SANE_Byte MaxMax = Max;
  SANE_Byte MinMin = Min;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_BSSCPowerDelay: start\n");

  *Target = (Max + Min) / 2;
  RIE((*pSetPowerDelay) (dev->chip, *Target));
  while (*Target != Min)
    {
      RIE(usb_high_scan_EvalueMaxLevel (dev, dev->INI_PowerDelayLines, Length,
					&MaxLevel));
      if (MaxLevel > Threshold)
	{
	  Min = *Target;
	  *Target = (Max + Min) / 2;
	  *ssCurrent = SG_BRIGHTER;
	}
      else if (MaxLevel < Threshold)
	{
	  Max = *Target;
	  *Target = (Max + Min) / 2;
	  *ssCurrent = SG_DARKER;
	}
      else if (MaxLevel == Threshold)
	{			/* Found. */
	  *ssCurrent = SG_EQUAL;
	  return SANE_STATUS_GOOD;
	}
      RIE((*pSetPowerDelay) (dev->chip, *Target));
    }
  /* Fail... */
  if (Max == MaxMax || Min == MinMin)
    {				/* Boundary check */
      if (Max == MaxMax)	/*Target on max side */
	*Target = MaxMax;
      else
	*Target = MinMin;
      RIE((*pSetPowerDelay) (dev->chip, *Target));
      RIE(usb_high_scan_EvalueMaxLevel (dev, dev->INI_PowerDelayLines, Length,
					&MaxLevel));

      if (MaxLevel > Threshold)
	{
	  *ssCurrent = SG_BRIGHTER;
	}
      else if (MaxLevel < Threshold)
	{
	  *ssCurrent = SG_DARKER;
	}
      else if (MaxLevel == Threshold)
	{
	  *ssCurrent = SG_EQUAL;
	}
    }
  else
    {/* Fail... will always on mimnum side, make it darker */
      Target++;
      *ssCurrent = SG_DARKER;
    }
  DBG (5, "usb_high_scan_BSSCPowerDelay: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_AdjustRgb600PowerDelay (Mustek_Usb_Device * dev)
{
  SANE_Status status;
  SANE_Byte MaxPowerDelay;
  SIGNALSTATE ssCurrent = SG_UNKNOW;
  
  DBG (5, "usb_high_scan_AdjustRgb600PowerDelay: start\n");
  MaxPowerDelay = (SANE_Byte) (dev->ExploseTime / 64);

  if (dev->isAdjustedRgb600PowerDelay)
    return SANE_STATUS_GOOD;
  /* Setup Initial State */
  dev->RedRgb600PowerDelay = MaxPowerDelay;
  dev->GreenRgb600PowerDelay = MaxPowerDelay;
  dev->BlueRgb600PowerDelay = MaxPowerDelay;

  RIE(usb_low_SetCCDWidth (dev->chip, dev->ExploseTime));
  RIE(usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd));
  RIE(usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef));
  RIE(usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset));
  RIE(usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset));
  RIE(usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset));
  RIE(usb_mid_front_SetRGBSignal (dev->chip));
  RIE(usb_low_SetDummy (dev->chip, dev->INI_600SkipsPerRow));
  RIE(usb_low_SetImageByteWidth (dev->chip, A4CIS_600AdjustLength));
  RIE(usb_low_SetPixelDeepth (dev->chip, PD_8BIT));

  /* Adjust GreenPD */
  RIE(usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN));
  if (dev->Sensor == SR_CANON300600)
    RIE(usb_mid_c300600_PrepareRgbX (dev->chip, 600));
  else if (dev->Sensor == SR_CANON300)
    RIE(usb_mid_c300_PrepareRgbX (dev->chip, 600)); /* fixme */
  else
    RIE(usb_mid_c600_PrepareRgbX (dev->chip, 600));
  ssCurrent = SG_UNKNOW;
  RIE(usb_mid_front_SetGreenPGA (dev->chip, dev->GreenRgb600PGA));
  RIE(usb_high_scan_BSSCPowerDelay 
      (dev, &usb_low_SetGreenPD, &ssCurrent, &dev->GreenRgb600PowerDelay,
     MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_600AdjustLength));

  /* Adjust BluePD */
  RIE(usb_mid_motor1200_PrepareAdjust (dev->chip, CH_BLUE));
  if (dev->Sensor == SR_CANON300600)
    RIE(usb_mid_c300600_PrepareRgbX (dev->chip, 600));
  else if (dev->Sensor == SR_CANON300)
    RIE(usb_mid_c300_PrepareRgbX (dev->chip, 600)); /* fixme */
  else
    RIE(usb_mid_c600_PrepareRgbX (dev->chip, 600));
  ssCurrent = SG_UNKNOW;
  RIE(usb_mid_front_SetBluePGA (dev->chip, dev->BlueRgb600PGA));
  RIE(usb_high_scan_BSSCPowerDelay 
    (dev, &usb_low_SetBluePD, &ssCurrent, &dev->BlueRgb600PowerDelay,
     MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_600AdjustLength));

  /* Adjust RedPD */
  RIE(usb_mid_motor1200_PrepareAdjust (dev->chip, CH_RED));
  if (dev->Sensor == SR_CANON300600)
    RIE(usb_mid_c300600_PrepareRgbX (dev->chip, 600));
  else if (dev->Sensor == SR_CANON300)
    RIE(usb_mid_c300_PrepareRgbX (dev->chip, 600)); /* fixme */
  else
    RIE(usb_mid_c600_PrepareRgbX (dev->chip, 600));
  ssCurrent = SG_UNKNOW;
  RIE(usb_mid_front_SetRedPGA (dev->chip, dev->RedRgb600PGA));
  RIE(usb_high_scan_BSSCPowerDelay 
    (dev, &usb_low_SetRedPD, &ssCurrent, &dev->RedRgb600PowerDelay,
     MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_600AdjustLength));

  dev->isAdjustedRgb600PowerDelay = SANE_TRUE;
  DBG (5, "usb_high_scan_AdjustRgb600PowerDelay: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_AdjustMono600PowerDelay (Mustek_Usb_Device * dev)
{
  SANE_Byte MaxPowerDelay;
  SIGNALSTATE ssCurrent = SG_UNKNOW;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_AdjustMono600PowerDelay: start\n");
  MaxPowerDelay = (SANE_Byte) (dev->ExploseTime / 64);
  if (dev->isAdjustedMono600PowerDelay)
    return SANE_STATUS_GOOD;
  /* Setup Initial State */
  dev->RedMono600PowerDelay = MaxPowerDelay;
  dev->GreenMono600PowerDelay = MaxPowerDelay;
  dev->BlueMono600PowerDelay = MaxPowerDelay;

  /* Compute Gray PD */
  RIE(usb_low_SetCCDWidth (dev->chip, dev->ExploseTime));
  RIE(usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd));
  RIE(usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef));
  RIE(usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset));
  RIE(usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset));
  RIE(usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset));
  RIE(usb_mid_front_SetRGBSignal (dev->chip));
  RIE(usb_low_SetDummy (dev->chip, dev->INI_600SkipsPerRow));
  RIE(usb_low_SetImageByteWidth (dev->chip, A4CIS_600AdjustLength));
  RIE(usb_low_SetPixelDeepth (dev->chip, PD_8BIT));

  /* Adjust GreenGrayPD */
  RIE(usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN));
  if (dev->Sensor == SR_CANON300600)
    RIE(usb_mid_c300600_PrepareRgbX (dev->chip, 600));
  else if (dev->Sensor == SR_CANON300)
    RIE(usb_mid_c300_PrepareRgbX (dev->chip, 600)); /* fixme */
  else
    RIE(usb_mid_c600_PrepareRgbX (dev->chip, 600));
  ssCurrent = SG_UNKNOW;
  RIE(usb_mid_front_SetRedPGA (dev->chip, dev->Mono600PGA));
  RIE(usb_mid_front_SetGreenPGA (dev->chip, dev->Mono600PGA));
  RIE(usb_mid_front_SetBluePGA (dev->chip, dev->Mono600PGA));
  RIE(usb_high_scan_BSSCPowerDelay 
    (dev, &usb_low_SetGreenPD, &ssCurrent, &dev->GreenMono600PowerDelay,
     MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_600AdjustLength));

  dev->isAdjustedMono600PowerDelay = SANE_TRUE;
  DBG (5, "usb_high_scan_AdjustMono600PowerDelay: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_AdjustRgb300PowerDelay (Mustek_Usb_Device * dev)
{
  /* Setup Initial State */
  SANE_Byte MaxPowerDelay;
  SIGNALSTATE ssCurrent = SG_UNKNOW;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_AdjustRgb300PowerDelay: start\n");
  MaxPowerDelay = (SANE_Byte) (dev->ExploseTime / 64);
  if (dev->isAdjustedRgb300PowerDelay)
    return SANE_STATUS_GOOD;

  dev->RedRgb300PowerDelay = MaxPowerDelay;
  dev->GreenRgb300PowerDelay = MaxPowerDelay;
  dev->BlueRgb300PowerDelay = MaxPowerDelay;

  RIE(usb_low_SetCCDWidth (dev->chip, dev->ExploseTime));
  RIE(usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd));
  RIE(usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef));
  RIE(usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset));
  RIE(usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset));
  RIE(usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset));
  RIE(usb_mid_front_SetRGBSignal (dev->chip));
  RIE(usb_low_SetDummy (dev->chip, dev->INI_300SkipsPerRow));
  RIE(usb_low_SetImageByteWidth (dev->chip, A4CIS_300AdjustLength));
  RIE(usb_low_SetPixelDeepth (dev->chip, PD_8BIT));

  /* Adjust GreenPD */
  RIE(usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN));
  if (dev->Sensor == SR_CANON300600)
    RIE(usb_mid_c300600_PrepareRgbX (dev->chip, 300));
  else if (dev->Sensor == SR_CANON300)
    RIE(usb_mid_c300_PrepareRgbX (dev->chip, 300)); /* fixme? */
  else
    RIE(usb_mid_c600_PrepareRgbX (dev->chip, 300));

  ssCurrent = SG_UNKNOW;
  RIE(usb_mid_front_SetGreenPGA (dev->chip,dev->GreenRgb300PGA));
  RIE(usb_high_scan_BSSCPowerDelay 
    (dev, &usb_low_SetGreenPD, &ssCurrent, &dev->GreenRgb300PowerDelay,
     MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength));

  /* Adjust BluePD */
  RIE(usb_mid_motor1200_PrepareAdjust (dev->chip, CH_BLUE));
  if (dev->Sensor == SR_CANON300600)
    RIE(usb_mid_c300600_PrepareRgbX (dev->chip, 300));
  else if (dev->Sensor == SR_CANON300)
    RIE(usb_mid_c300_PrepareRgbX (dev->chip, 300)); /* fixme */
  else
    RIE(usb_mid_c600_PrepareRgbX (dev->chip, 300));

  ssCurrent = SG_UNKNOW;
  RIE(usb_mid_front_SetBluePGA (dev->chip, dev->BlueRgb300PGA));
  RIE(usb_high_scan_BSSCPowerDelay
    (dev, &usb_low_SetBluePD, &ssCurrent, &dev->BlueRgb300PowerDelay,
     MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength));

  /* Adjust RedPD */
  RIE(usb_mid_motor1200_PrepareAdjust (dev->chip, CH_RED));
  if (dev->Sensor == SR_CANON300600)
    RIE(usb_mid_c300600_PrepareRgbX (dev->chip, 300));
  else if (dev->Sensor == SR_CANON300)
    RIE(usb_mid_c300_PrepareRgbX (dev->chip, 300)); /* fixme */
  else
    RIE(usb_mid_c600_PrepareRgbX (dev->chip, 300));

  ssCurrent = SG_UNKNOW;
  RIE(usb_mid_front_SetRedPGA (dev->chip, dev->RedRgb300PGA));
  RIE(usb_high_scan_BSSCPowerDelay
    (dev, &usb_low_SetRedPD, &ssCurrent, &dev->RedRgb300PowerDelay,
     MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength));
  dev->isAdjustedRgb300PowerDelay = SANE_TRUE;
  DBG (5, "usb_high_scan_AdjustRgb300PowerDelay: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_AdjustMono300PowerDelay (Mustek_Usb_Device * dev)
{
  SANE_Byte MaxPowerDelay;
  SIGNALSTATE ssCurrent = SG_UNKNOW;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_AdjustMono300PowerDelay: start\n");
  MaxPowerDelay = (SANE_Byte) (dev->ExploseTime / 64);
  if (dev->isAdjustedMono300PowerDelay)
    return SANE_STATUS_GOOD;
  /* Setup Initial State */
  dev->RedMono300PowerDelay = MaxPowerDelay;
  dev->GreenMono300PowerDelay = MaxPowerDelay;
  dev->BlueMono300PowerDelay = MaxPowerDelay;

  /* Compute Gray PD */
  RIE(usb_low_SetCCDWidth (dev->chip, dev->ExploseTime));
  RIE(usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd));
  RIE(usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef));
  RIE(usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset));
  RIE(usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset));
  RIE(usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset));
  RIE(usb_mid_front_SetRGBSignal (dev->chip));
  RIE(usb_low_SetDummy (dev->chip, dev->INI_300SkipsPerRow));
  RIE(usb_low_SetImageByteWidth (dev->chip, A4CIS_300AdjustLength));
  RIE(usb_low_SetPixelDeepth (dev->chip, PD_8BIT));

  /* Adjust GreenGrayPD */
  RIE(usb_mid_motor1200_PrepareAdjust (dev->chip, CH_GREEN));

  if (dev->Sensor == SR_CANON300600)
    RIE(usb_mid_c300600_PrepareRgbX (dev->chip, 300));
  else if (dev->Sensor == SR_CANON300)
    RIE(usb_mid_c300_PrepareRgbX (dev->chip, 300)); /* fixme */
  else
    RIE(usb_mid_c600_PrepareRgbX (dev->chip, 300));

  ssCurrent = SG_UNKNOW;
  RIE(usb_mid_front_SetRedPGA (dev->chip, dev->Mono300PGA));
  RIE(usb_mid_front_SetGreenPGA (dev->chip, dev->Mono300PGA));
  RIE(usb_mid_front_SetBluePGA (dev->chip, dev->Mono300PGA));
  RIE(usb_high_scan_BSSCPowerDelay
    (dev, &usb_low_SetGreenPD, &ssCurrent, &dev->GreenMono300PowerDelay,
     MaxPowerDelay, 0, dev->INI_MaxPowerDelay, A4CIS_300AdjustLength));

  dev->isAdjustedMono300PowerDelay = SANE_TRUE;
  DBG (5, "usb_high_scan_AdjustMono300PowerDelay: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_EvaluePixelRate (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_EvaluePixelRate: start, dev=%p\n", dev);
  dev->PixelRate = 2000;
  dev->isEvaluePixelRate = SANE_TRUE;
  DBG (5, "usb_high_scan_EvaluePixelRate: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_CalibrationRgb24Bit (Mustek_Usb_Device * dev)
{
  SANE_Word WhiteNeed;
  SANE_Word DarkNeed;
  SANE_Word i;
  SANE_Status status;
  SANE_Word lines_left;

  DBG (5, "usb_high_scan_CalibrationRgb24Bit: start, dev=%p\n", dev);
  RIE(usb_mid_motor1200_PrepareCalibrateRgbY (dev->chip, dev->YDpi));
  RIE(usb_low_TurnLampPower (dev->chip, SANE_TRUE));

  dev->pRedCalibrator = (CalMD12 *) malloc (sizeof (CalMD12));
  if (!dev->pRedCalibrator)
    return SANE_STATUS_NO_MEM;

  RIE(usb_high_cal_MD12_init (dev->pRedCalibrator, I8O8RGB, 
			      dev->INI_KLevel << 8, 0));
  RIE(usb_high_cal_MD12_Prepare (dev->pRedCalibrator, dev->Width));
  RIE(usb_high_cal_MD12_EmbedGamma (dev->pRedCalibrator, dev->pGammaTable));
  RIE(usb_high_cal_MD12_Setup (dev->pRedCalibrator, 1, 2, 8, dev->Width,
			       &WhiteNeed, &DarkNeed));

  dev->pGreenCalibrator =
    (CalMD12 *) malloc (sizeof (CalMD12));
  if (!dev->pGreenCalibrator)
    return SANE_STATUS_NO_MEM;
  RIE(usb_high_cal_MD12_init (dev->pGreenCalibrator, I8O8RGB,
			      dev->INI_KLevel << 8, 0));
  RIE(usb_high_cal_MD12_Prepare (dev->pGreenCalibrator, dev->Width));
  RIE(usb_high_cal_MD12_EmbedGamma (dev->pGreenCalibrator, dev->pGammaTable));
  RIE(usb_high_cal_MD12_Setup (dev->pGreenCalibrator, 1, 2, 8, dev->Width,
			       &WhiteNeed, &DarkNeed));

  dev->pBlueCalibrator = (CalMD12 *) malloc (sizeof (CalMD12));
  RIE(usb_high_cal_MD12_init (dev->pBlueCalibrator, I8O8RGB,
			      dev->INI_KLevel << 8, 0));
  RIE(usb_high_cal_MD12_Prepare (dev->pBlueCalibrator, dev->Width));
  RIE(usb_high_cal_MD12_EmbedGamma (dev->pBlueCalibrator, dev->pGammaTable));
  RIE(usb_high_cal_MD12_Setup (dev->pBlueCalibrator, 1, 2, 8, dev->Width,
			       &WhiteNeed, &DarkNeed));

  /* K White */
  RIE(usb_low_StartRowing (dev->chip));
  for (i = 0; i < WhiteNeed; i++)
    {
      /* Read Green Channel */
      RIE(usb_low_GetRow (dev->chip, dev->pGreen, &lines_left));
      RIE(usb_high_cal_MD12_FillInWhite (dev->pGreenCalibrator, i, 0,
					 (void *) (dev->pGreen +
						   dev->SkipsPerRow)));
      RIE(usb_low_GetRow (dev->chip, dev->pGreen, &lines_left));
      RIE(usb_high_cal_MD12_FillInWhite (dev->pGreenCalibrator, i, 1,
					 (void *) (dev->pGreen +
						   dev->SkipsPerRow)));
      /* Read Blue Channel */
      RIE(usb_low_GetRow (dev->chip, dev->pBlue, &lines_left));
      RIE(usb_high_cal_MD12_FillInWhite (dev->pBlueCalibrator, i, 0,
					 (void *) (dev->pBlue +
						   dev->SkipsPerRow)));
      RIE(usb_low_GetRow (dev->chip, dev->pBlue, &lines_left));
      RIE(usb_high_cal_MD12_FillInWhite (dev->pBlueCalibrator, i, 1,
					 (void *) (dev->pBlue +
						   dev->SkipsPerRow)));
      /* Read Red Channel */
      RIE(usb_low_GetRow (dev->chip, dev->pRed, &lines_left));
      RIE(usb_high_cal_MD12_FillInWhite (dev->pRedCalibrator, i, 0,
					 (void *) (dev->pRed +
						   dev->SkipsPerRow)));
      RIE(usb_low_GetRow (dev->chip, dev->pRed, &lines_left));
      RIE(usb_high_cal_MD12_FillInWhite (dev->pRedCalibrator, i, 1,
					 (void *) (dev->pRed +
						   dev->SkipsPerRow)));
    }
  RIE(usb_low_StopRowing (dev->chip));
  /* Caculate Aver */
  RIE(usb_high_cal_MD12_EvalueWhite (dev->pGreenCalibrator,
				     dev->INI_GreenFactor));
  RIE(usb_high_cal_MD12_EvalueWhite (dev->pBlueCalibrator, 
				     dev->INI_BlueFactor));
  RIE(usb_high_cal_MD12_EvalueWhite (dev->pRedCalibrator,
				     dev->INI_RedFactor));

  RIE(usb_mid_motor1200_PrepareCalibrateRgbY (dev->chip, dev->YDpi));
  RIE(usb_low_EnableMotor (dev->chip, SANE_FALSE));
  RIE(usb_low_TurnLampPower (dev->chip, SANE_FALSE));

  /* K Black */
  RIE(usb_low_StartRowing (dev->chip));
  for (i = 0; i < DarkNeed; i++)
    {
      /* Read Green Channel */
      RIE(usb_low_GetRow (dev->chip, dev->pGreen, &lines_left));
      RIE(usb_high_cal_MD12_FillInDark (dev->pGreenCalibrator, i, 0,
					(void *) (dev->pGreen +
						  dev->SkipsPerRow)));
      RIE(usb_low_GetRow (dev->chip, dev->pGreen, &lines_left));
      RIE(usb_high_cal_MD12_FillInDark (dev->pGreenCalibrator, i, 1,
					(void *) (dev->pGreen +
						  dev->SkipsPerRow)));
      /* Read Blue Channel */
      RIE(usb_low_GetRow (dev->chip, dev->pBlue, &lines_left));
      RIE(usb_high_cal_MD12_FillInDark (dev->pBlueCalibrator, i, 0,
					(void *) (dev->pBlue +
						  dev->SkipsPerRow)));
      RIE(usb_low_GetRow (dev->chip, dev->pBlue, &lines_left));
      RIE(usb_high_cal_MD12_FillInDark (dev->pBlueCalibrator, i, 1,
					(void *) (dev->pBlue +
						  dev->SkipsPerRow)));
      /* Read Red Channel */
      RIE(usb_low_GetRow (dev->chip, dev->pRed, &lines_left));
      RIE(usb_high_cal_MD12_FillInDark (dev->pRedCalibrator, i, 0,
					(void *) (dev->pRed +
						  dev->SkipsPerRow)));
      RIE(usb_low_GetRow (dev->chip, dev->pRed, &lines_left));
      RIE(usb_high_cal_MD12_FillInDark (dev->pRedCalibrator, i, 1,
					(void *) (dev->pRed +
						  dev->SkipsPerRow)));
    }
  RIE(usb_low_StopRowing (dev->chip));
  RIE(usb_low_TurnLampPower (dev->chip, SANE_TRUE));
  /* Caculate Aver */
  RIE(usb_high_cal_MD12_EvalueDark (dev->pGreenCalibrator,
				    dev->INI_GreenBlackFactor));
  RIE(usb_high_cal_MD12_EvalueDark (dev->pBlueCalibrator,
				    dev->INI_BlueBlackFactor));
  RIE(usb_high_cal_MD12_EvalueDark (dev->pRedCalibrator,
				    dev->INI_RedBlackFactor));
  /* Caculate Mapping */
  RIE(usb_high_cal_MD12_EvalueCalibrator (dev->pGreenCalibrator));
  RIE(usb_high_cal_MD12_EvalueCalibrator (dev->pBlueCalibrator));
  RIE(usb_high_cal_MD12_EvalueCalibrator (dev->pRedCalibrator));
  DBG (5, "usb_high_scan_CalibrationRgb24Bit: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_CalibrationMono8Bit (Mustek_Usb_Device * dev)
{
  SANE_Word WhiteNeed;
  SANE_Word DarkNeed;
  SANE_Word i;
  SANE_Status status;
  SANE_Word lines_left;
  
  DBG (5, "usb_high_scan_CalibrationMono8Bit: start\n");
  RIE(usb_mid_motor1200_PrepareCalibrateMonoY (dev->chip, dev->YDpi));
  RIE(usb_low_TurnLampPower (dev->chip, SANE_TRUE));

  dev->pMonoCalibrator = (CalMD12 *) malloc (sizeof (CalMD12));
  RIE(usb_high_cal_MD12_init (dev->pMonoCalibrator, I8O8MONO, 
			      dev->INI_KLevel << 8, 0));
  RIE(usb_high_cal_MD12_Prepare (dev->pMonoCalibrator, dev->Width));
  RIE(usb_high_cal_MD12_EmbedGamma (dev->pMonoCalibrator,
				    dev->pGammaTable));
  RIE(usb_high_cal_MD12_Setup (dev->pMonoCalibrator, 1, 1, 8,
			       dev->Width, &WhiteNeed, &DarkNeed));

  /* K White */
  RIE(usb_low_StartRowing (dev->chip));
  for (i = 0; i < WhiteNeed; i++)
    {
      /* Read Green Channel */
      RIE(usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine, &lines_left));
      RIE(usb_high_cal_MD12_FillInWhite (dev->pMonoCalibrator, i, 0,
					 (void *) (dev->A4CIS_GrayLine +
						   dev->SkipsPerRow)));
    }
  RIE(usb_low_StopRowing (dev->chip));
  /* Caculate Aver */
  RIE(usb_high_cal_MD12_EvalueWhite (dev->pMonoCalibrator, 
				     dev->INI_GrayFactor));

  RIE(usb_mid_motor1200_PrepareCalibrateMonoY (dev->chip, dev->YDpi));
  RIE(usb_low_EnableMotor (dev->chip, SANE_FALSE));
  RIE(usb_low_TurnLampPower (dev->chip, SANE_FALSE));

  /* K Black */
  RIE(usb_low_StartRowing (dev->chip));
  for (i = 0; i < DarkNeed; i++)
    {
      /* Read Green Channel */
      RIE(usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine, &lines_left));
      RIE(usb_high_cal_MD12_FillInDark (dev->pMonoCalibrator, i, 0,
					(void *) (dev->A4CIS_GrayLine +
						  dev->SkipsPerRow)));
    }
  RIE(usb_low_StopRowing (dev->chip));
  RIE(usb_low_TurnLampPower (dev->chip, SANE_TRUE));
  /* Caculate Green Black */
  RIE(usb_high_cal_MD12_EvalueDark (dev->pMonoCalibrator,
				    dev->INI_GrayBlackFactor));
  /* Caculate Mapping */
  RIE(usb_high_cal_MD12_EvalueCalibrator (dev->pMonoCalibrator));
  DBG (5, "usb_high_scan_CalibrationMono8Bit: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_StepForward (Mustek_Usb_Device * dev, int StepCount)
{
  SANE_Status status;

  DBG (5, "usb_high_scan_StepForward: start\n");
  
  if (StepCount <= 0)
    return SANE_STATUS_INVAL;
  /* Initialize */
  RIE(usb_low_SetCCDWidth (dev->chip, dev->INI_MinExploseTime));
  RIE(usb_low_SetMotorDirection (dev->chip, SANE_FALSE));
  if (dev->chip->scanner_type == mustek_usb_type_600cu)
    RIE(usb_mid_motor600_PrepareStep (dev->chip, (SANE_Word) StepCount));
  else
    RIE(usb_mid_motor1200_PrepareStep (dev->chip, (SANE_Word) StepCount));
  /* Startup */
  RIE(usb_low_StartRowing (dev->chip));
  /* Wait to Stop */
  /* Linux USB seems buggy on timeout... sleep before really try  */
  /* to read the flag from scanner */
  usleep (StepCount * 2 * 1000);
  RIE(usb_low_WaitRowingStop (dev->chip));
  RIE(usb_low_SetCCDWidth (dev->chip, dev->ExploseTime));

  DBG (5, "usb_high_scan_StepForward: start\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_SafeForward (Mustek_Usb_Device * dev, int StepCount)
{
  SANE_Status status;
  
  DBG (5, "usb_high_scan_SafeForward: start\n");
  if (StepCount <= 0)
    return SANE_STATUS_INVAL;
  /* Initialize */
  RIE(usb_low_SetCCDWidth (dev->chip, 5400));
  RIE(usb_low_SetMotorDirection (dev->chip, SANE_FALSE));
  if (dev->chip->scanner_type == mustek_usb_type_600cu)
    RIE(usb_mid_motor600_PrepareStep (dev->chip, (SANE_Word) StepCount));
  else
    RIE(usb_mid_motor1200_PrepareStep (dev->chip, (SANE_Word) StepCount));
  /* Startup */
  RIE(usb_low_StartRowing (dev->chip));
  /* Wait to Stop */
  RIE(usb_low_WaitRowingStop (dev->chip));
  RIE(usb_low_SetCCDWidth (dev->chip, dev->ExploseTime));
  DBG (5, "usb_high_scan_SafeForward: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Word
usb_high_scan_CaculateMaxRgb600Explose (Mustek_Usb_Device * dev,
					SANE_Byte * IdealRedPD,
					SANE_Byte * IdealGreenPD,
					SANE_Byte * IdealBluePD)
{
  SANE_Word RedLightUp;
  SANE_Word GreenLightUp;
  SANE_Word BlueLightUp;
  SANE_Word MaxLightUp;
  SANE_Word IdealExploseTime;

  DBG (5, "usb_high_scan_CaculateMaxRgb600Explose: dev=%p\n", dev);

  RedLightUp = dev->ExploseTime - dev->RedRgb600PowerDelay * 64;
  GreenLightUp = dev->ExploseTime - dev->GreenRgb600PowerDelay * 64;
  BlueLightUp = dev->ExploseTime - dev->BlueRgb600PowerDelay * 64;
  MaxLightUp = max (RedLightUp, max (GreenLightUp, BlueLightUp));
  IdealExploseTime = max (max (5376, MaxLightUp),
			  usb_mid_motor1200_RgbCapability (dev->YDpi));


  IdealExploseTime = (IdealExploseTime + 63) / 64 * 64;
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime - RedLightUp) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - GreenLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime - BlueLightUp) / 64);
  DBG (5, "usb_high_scan_CaculateMaxRgb600Explose: exit\n");
  return IdealExploseTime;
}

SANE_Word
usb_high_scan_CaculateMaxMono600Explose (Mustek_Usb_Device * dev, 
					 SANE_Byte * IdealRedPD,
					 SANE_Byte * IdealGreenPD,
					 SANE_Byte * IdealBluePD)
{
  SANE_Word MaxLightUp;
  SANE_Word IdealExploseTime;
  SANE_Word TransferTime;

  DBG (5, "usb_high_scan_CaculateMaxMono600Explose: dev=%p\n", dev);

  MaxLightUp = dev->ExploseTime - dev->GreenMono600PowerDelay * 64;
  TransferTime =  (SANE_Word) ((SANE_Word) (dev->PixelRate) 
			       * (SANE_Word) (dev->XDpi) / 600);
  /* base on 600, but double it. */

  if (TransferTime > 16000)
    TransferTime = 16000;
  IdealExploseTime 
    = max (max (5376, MaxLightUp), 
	   max (TransferTime, usb_mid_motor1200_MonoCapability (dev->YDpi)));
  IdealExploseTime = (IdealExploseTime + 63) / 64 * 64;
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - MaxLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime) / 64);
  DBG (5, "usb_high_scan_CaculateMaxMono600Explose: exit\n");
  return IdealExploseTime;
}

SANE_Word
usb_high_scan_CaculateMaxRgb300Explose (Mustek_Usb_Device * dev, 
					SANE_Byte * IdealRedPD,
					SANE_Byte * IdealGreenPD,
					SANE_Byte * IdealBluePD)
{
  SANE_Word RedLightUp;
  SANE_Word GreenLightUp;
  SANE_Word BlueLightUp;
  SANE_Word MaxLightUp;
  SANE_Word IdealExploseTime;

  DBG (5, "usb_high_scan_CaculateMaxRgb300Explose: start\n");
  RedLightUp = dev->ExploseTime - dev->RedRgb300PowerDelay * 64;
  GreenLightUp = dev->ExploseTime - dev->GreenRgb300PowerDelay * 64;
  BlueLightUp = dev->ExploseTime - dev->BlueRgb300PowerDelay * 64;
  MaxLightUp = max (RedLightUp, max (GreenLightUp, BlueLightUp));

  if (dev->Sensor == SR_CANON300600)
    {
      IdealExploseTime =
	max (max (2624, MaxLightUp),
	     usb_mid_motor1200_RgbCapability (dev->YDpi));
    }
  else if (dev->Sensor == SR_CANON300)
    {
      IdealExploseTime =
	max (max (2624, MaxLightUp), /* fixme? */
	     usb_mid_motor1200_RgbCapability (dev->YDpi));
    }
  else
    {
      IdealExploseTime =
	max (max (5376, MaxLightUp),
	     usb_mid_motor1200_RgbCapability (dev->YDpi));
    }

  IdealExploseTime = (IdealExploseTime + 63) / 64 * 64;
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime - RedLightUp) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - GreenLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime - BlueLightUp) / 64);
  DBG (5, "usb_high_scan_CaculateMaxRgb300Explose: exit\n");
  return IdealExploseTime;
}

SANE_Word
usb_high_scan_CaculateMaxMono300Explose (Mustek_Usb_Device * dev,
					 SANE_Byte * IdealRedPD,
					 SANE_Byte * IdealGreenPD,
					 SANE_Byte * IdealBluePD)
{
  SANE_Word MaxLightUp;
  SANE_Word TransferTime;
  SANE_Word IdealExploseTime;

  DBG (5, "usb_high_scan_CaculateMaxMono300Explose: start\n");
  MaxLightUp = dev->ExploseTime - dev->GreenMono300PowerDelay * 64;
  TransferTime =
    (SANE_Word) ((SANE_Word) (dev->PixelRate) * 
		 (SANE_Word) (dev->XDpi) / 600);
  /* base on 600, but double it. */

  if (TransferTime > 16000)
    TransferTime = 16000;
  if (dev->Sensor == SR_CANON300600)
    {
      IdealExploseTime =
	max (max (2688, MaxLightUp),
	     max (TransferTime,
		  usb_mid_motor1200_MonoCapability (dev->YDpi)));
    }
  else if (dev->Sensor == SR_CANON300)
    {
      IdealExploseTime =
	max (max (2688, MaxLightUp), /* fixme? */
	     max (TransferTime,
		  usb_mid_motor1200_MonoCapability (dev->YDpi)));
    }
  else
    {
      IdealExploseTime =
	max (max (5376, MaxLightUp),
	     max (TransferTime,
		  usb_mid_motor1200_MonoCapability (dev->YDpi)));
    }

  IdealExploseTime = (IdealExploseTime + 63) / 64 * 64;
  *IdealRedPD = (SANE_Byte) ((IdealExploseTime) / 64);
  *IdealGreenPD = (SANE_Byte) ((IdealExploseTime - MaxLightUp) / 64);
  *IdealBluePD = (SANE_Byte) ((IdealExploseTime) / 64);
  DBG (5, "usb_high_scan_CaculateMaxMono300Explose: exit\n");
  return IdealExploseTime;
}

SANE_Status
usb_high_scan_PrepareRgbSignal600Dpi (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  SANE_Byte IdealRedPD, IdealGreenPD, IdealBluePD;
  SANE_Word IdealExploseTime;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_PrepareRgbSignal600Dpi: start\n");
  IdealExploseTime =
    usb_high_scan_CaculateMaxRgb600Explose (dev, &IdealRedPD, &IdealGreenPD,
					    &IdealBluePD);

  RIE(usb_low_SetCCDWidth (dev->chip, IdealExploseTime));
  RIE(usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd));
  RIE(usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef));
  RIE(usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset));
  RIE(usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset));
  RIE(usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset));
  RIE(usb_mid_front_SetRedPGA (dev->chip, dev->RedRgb600PGA));
  RIE(usb_mid_front_SetGreenPGA (dev->chip, dev->GreenRgb600PGA));
  RIE(usb_mid_front_SetBluePGA (dev->chip, dev->BlueRgb600PGA));
  RIE(usb_mid_front_SetRGBSignal (dev->chip));
  RIE(usb_low_SetRedPD (dev->chip, IdealRedPD));
  RIE(usb_low_SetGreenPD (dev->chip, IdealGreenPD));
  RIE(usb_low_SetBluePD (dev->chip, IdealBluePD));
  DBG (5, "usb_high_scan_PrepareRgbSignal600Dpi: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_PrepareMonoSignal600Dpi (Mustek_Usb_Device * dev)
{
  SANE_Byte IdealRedPD, IdealGreenPD, IdealBluePD;
  SANE_Word IdealExploseTime;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_PrepareMonoSignal600Dpi: start\n");
  IdealExploseTime =
    usb_high_scan_CaculateMaxMono600Explose (dev, &IdealRedPD, &IdealGreenPD,
					     &IdealBluePD);

  RIE(usb_low_SetCCDWidth (dev->chip, IdealExploseTime));
  RIE(usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd));
  RIE(usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef));
  RIE(usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset));
  RIE(usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset));
  RIE(usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset));
  RIE(usb_mid_front_SetRedPGA (dev->chip, dev->Mono600PGA));
  RIE(usb_mid_front_SetGreenPGA (dev->chip, dev->Mono600PGA));
  RIE(usb_mid_front_SetBluePGA (dev->chip, dev->Mono600PGA));
  RIE(usb_mid_front_SetRGBSignal (dev->chip));
  RIE(usb_low_SetRedPD (dev->chip, IdealRedPD));
  RIE(usb_low_SetGreenPD (dev->chip, IdealGreenPD));
  RIE(usb_low_SetBluePD (dev->chip, IdealBluePD));
  DBG (5, "usb_high_scan_PrepareMonoSignal600Dpi: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_PrepareRgbSignal300Dpi (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  SANE_Byte IdealRedPD, IdealGreenPD, IdealBluePD;
  SANE_Word IdealExploseTime;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_PrepareRgbSignal300Dpi: start\n");

  IdealExploseTime =
    usb_high_scan_CaculateMaxRgb300Explose (dev, &IdealRedPD, &IdealGreenPD,
					&IdealBluePD);

  RIE(usb_low_SetCCDWidth (dev->chip, IdealExploseTime));
  RIE(usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd));
  RIE(usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef));
  RIE(usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset));
  RIE(usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset));
  RIE(usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset));
  RIE(usb_mid_front_SetRedPGA (dev->chip, dev->RedRgb300PGA));
  RIE(usb_mid_front_SetGreenPGA (dev->chip, dev->GreenRgb300PGA));
  RIE(usb_mid_front_SetBluePGA (dev->chip, dev->BlueRgb300PGA));
  RIE(usb_mid_front_SetRGBSignal (dev->chip));
  RIE(usb_low_SetRedPD (dev->chip, IdealRedPD));
  RIE(usb_low_SetGreenPD (dev->chip, IdealGreenPD));
  RIE(usb_low_SetBluePD (dev->chip, IdealBluePD));
  DBG (5, "usb_high_scan_PrepareRgbSignal300Dpi: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_PrepareMonoSignal300Dpi (Mustek_Usb_Device * dev)
{
  /* Setup Scan Here */
  SANE_Byte IdealRedPD, IdealGreenPD, IdealBluePD;
  SANE_Word IdealExploseTime;
  SANE_Status status;
  
  DBG (5, "usb_high_scan_PrepareMonoSignal300Dpi: start\n");
  IdealExploseTime =
    usb_high_scan_CaculateMaxMono300Explose (dev, &IdealRedPD, &IdealGreenPD,
					     &IdealBluePD);

  RIE(usb_low_SetCCDWidth (dev->chip, IdealExploseTime));
  RIE(usb_mid_front_SetFrontEndMode (dev->chip, dev->INI_FrontEnd));
  RIE(usb_mid_front_SetTopReference (dev->chip, dev->INI_TopRef));
  RIE(usb_mid_front_SetRedOffset (dev->chip, dev->INI_RedOffset));
  RIE(usb_mid_front_SetGreenOffset (dev->chip, dev->INI_GreenOffset));
  RIE(usb_mid_front_SetBlueOffset (dev->chip, dev->INI_BlueOffset));
  RIE(usb_mid_front_SetRedPGA (dev->chip, dev->Mono300PGA));
  RIE(usb_mid_front_SetGreenPGA (dev->chip, dev->Mono300PGA));
  RIE(usb_mid_front_SetBluePGA (dev->chip, dev->Mono300PGA));
  RIE(usb_mid_front_SetRGBSignal (dev->chip));
  RIE(usb_low_SetRedPD (dev->chip, IdealRedPD));
  RIE(usb_low_SetGreenPD (dev->chip, IdealGreenPD));
  RIE(usb_low_SetBluePD (dev->chip, IdealBluePD));
  DBG (5, "usb_high_scan_PrepareMonoSignal300Dpi: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_PrepareRgb24Bit (Mustek_Usb_Device * dev)
{
  SANE_Status status;
  
  DBG (5, "usb_high_scan_PrepareRgb24Bit: start\n");
  RIE(usb_low_SetImageByteWidth (dev->chip, dev->BytesPerStrip));
  RIE(usb_low_SetDummy (dev->chip, dev->Dummy));
  RIE(usb_low_SetPixelDeepth (dev->chip, PD_8BIT));
  DBG (5, "usb_high_scan_PrepareRgb24Bit: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_PrepareMono8Bit (Mustek_Usb_Device * dev)
{
  SANE_Status status;
  
  DBG (5, "usb_high_scan_PrepareMono8Bit: start\n");
  RIE(usb_low_SetImageByteWidth (dev->chip, dev->BytesPerStrip));
  RIE(usb_low_SetDummy (dev->chip, dev->Dummy));
  RIE(usb_low_SetPixelDeepth (dev->chip, PD_8BIT));
  DBG (5, "usb_high_scan_PrepareMono8Bit: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_GetRgb24BitLine (Mustek_Usb_Device * dev, SANE_Byte * lpLine,
			       SANE_Bool isOrderInvert)
{
  SANE_Status status;
  SANE_Word lines_left;
  
  DBG (5, "usb_high_scan_GetRgb24BitLine: start, dev=%p, lpLine=%p, "
       "isOrderInvert=%d\n", dev, lpLine, isOrderInvert);

  RIE(usb_low_GetRow (dev->chip, dev->pGreen, &lines_left));

  RIE(usb_low_GetRow (dev->chip, dev->pBlue, &lines_left));

  RIE(usb_low_GetRow (dev->chip, dev->pRed, &lines_left));
  RIE(usb_high_cal_MD12_Calibrate (dev->pGreenCalibrator,
				   dev->pGreen + dev->SkipsPerRow,
				   lpLine + 1));
  RIE(usb_high_cal_MD12_Calibrate (dev->pBlueCalibrator,
				   dev->pBlue + dev->SkipsPerRow,
				   lpLine + ((isOrderInvert) ? 0 : 2)));
  RIE(usb_high_cal_MD12_Calibrate (dev->pRedCalibrator,
				   dev->pRed + dev->SkipsPerRow,
				   lpLine + ((isOrderInvert) ? 2 : 0)));

  DBG (5, "usb_high_scan_GetRgb24BitLine: exit\n");
  return SANE_STATUS_GOOD;
}


SANE_Status
usb_high_scan_GetMono8BitLine (Mustek_Usb_Device * dev, SANE_Byte * lpLine,
			       SANE_Bool isOrderInvert)
{
  SANE_Status status;
  SANE_Word lines_left;

  DBG (5, "usb_high_scan_GetMono8BitLine: start, dev=%p, lpLine=%p, "
       "isOrderInvert=%d\n", dev, lpLine, isOrderInvert);

  RIE(usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine, &lines_left));
  RIE(usb_high_cal_MD12_Calibrate (dev->pMonoCalibrator,
				   dev->A4CIS_GrayLine +
				   dev->SkipsPerRow,
				   lpLine));
  DBG (5, "usb_high_scan_GetMono8BitLine: exit\n");
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_BackTrackRgb24Bit (Mustek_Usb_Device * dev)
{
  DBG (5, "usb_high_scan_BackTrackRgb24Bit: noop, dev=%p\n", dev);
  return SANE_STATUS_GOOD;
}

SANE_Status
usb_high_scan_BackTrackMono8Bit (Mustek_Usb_Device * dev)
{
  int i;
  SANE_Status status;
  SANE_Word lines_left;
  
  DBG (5, "usb_high_scan_BackTrackMono8Bit: start, dev=%p\n", dev);

  if (dev->YDpi >= 300)
    {
      RIE(usb_low_StopRowing (dev->chip));
      RIE(usb_low_SetMotorDirection (dev->chip, SANE_TRUE));
      RIE(usb_low_StartRowing (dev->chip));
      for (i = 0; i < dev->INI_Mono8BitBackTrack; i++)
	{
	  RIE(usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine, &lines_left));
	}
      usleep (100 * 1000);
      RIE(usb_low_StopRowing (dev->chip));
      RIE(usb_low_SetMotorDirection (dev->chip, SANE_FALSE));
      RIE(usb_low_StartRowing (dev->chip));
      for (i = 0; i < dev->INI_Mono8BitBackTrack; i++)
	{
	  RIE(usb_low_GetRow (dev->chip, dev->A4CIS_GrayLine, &lines_left));
	}
    }
  DBG (5, "usb_high_scan_BackTrackMono8Bit: exit\n");
  return SANE_STATUS_GOOD;
}
